from __future__ import annotations

"""
Type Checking Support 
"""

import typing as t
from lzl import load

if load.TYPE_CHECKING:
    import kr8s
else:
    kr8s = load.LazyLoad("kr8s", install_missing=True)

if t.TYPE_CHECKING:
    from kr8s import objects
    from kr8s.asyncio import objects as aobjects
    ObjectT = t.Union[
        objects.Pod,
        objects.Service,
        objects.Deployment,
        objects.StatefulSet,
        objects.ConfigMap,
        objects.Secret,
        objects.PersistentVolume,
        objects.PersistentVolumeClaim,
        objects.Namespace,
        objects.Node,
        objects.Event,
        objects.Endpoints,
        objects.Ingress,
        objects.ServiceAccount,
        objects.Role,
        objects.RoleBinding,
        objects.ClusterRole,
        objects.APIObject,
    ]
    ObjectListT = t.List[ObjectT]
    aObjectT = t.Union[
        aobjects.Pod,
        aobjects.Service,
        aobjects.Deployment,
        aobjects.StatefulSet,
        aobjects.ConfigMap,
        aobjects.Secret,
        aobjects.PersistentVolume,
        aobjects.PersistentVolumeClaim,
        aobjects.Namespace,
        aobjects.Node,
        aobjects.Event,
        aobjects.Endpoints,
        aobjects.Ingress,
        aobjects.ServiceAccount,
        aobjects.Role,
        aobjects.RoleBinding,
        aobjects.ClusterRole,
        aobjects.APIObject,
    ]
    aObjectListT = t.List[aObjectT]




from __future__ import annotations

import random
from lzl.api import aiohttpx
from lzl.types import BaseModel, Field
from typing import Optional, Dict, Union, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..configs.openai import OpenAISettings, AzureOpenAISettings
    from ..configs.external import ExternalProviderConfig

class OpenAIContext(BaseModel):
    """
    A context object for OpenAI
    """
    custom_headers: Optional[Dict[str, str]] = Field(default_factory=dict)
    data: Optional[Dict[str, Any]] = Field(default_factory=dict)

    def update_headers(self, headers: Dict[str, Any]):
        """
        Updates the custom headers
        """
        if self.custom_headers is None: self.custom_headers = {}
        for k,v in headers.items():
            if isinstance(v, str): continue
            if isinstance(v, bool):
                headers[k] = "true" if v else "false"
            if isinstance(v, (int, float, type(None))):
                headers[k] = str(v)
        self.custom_headers.update(headers)


class OpenAIAuth(aiohttpx.Auth):
    """
    Custom Authentication Wrapper for OpenAI Client
    """
    def __init__(
        self, 
        settings: Union['OpenAISettings', 'AzureOpenAISettings'],
        context: 'OpenAIContext',
        auth_key: Optional[str] = None,
        auth_value: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the OpenAI Auth
        """
        self.settings = settings
        self.context = context
        self.auth_key = auth_key
        self.auth_value = auth_value
    
    def auth_flow(self, request: aiohttpx.Request):
        """
        Injects the API Key into the Request
        """
        # request.headers.update(self.settings.get_api_key_headers())
        if self.auth_key not in request.headers:
            request.headers[self.auth_key] = self.auth_value
        if custom_headers := self.context.custom_headers:
            request.headers.update(custom_headers)
        
        # if self.settings.custom_headers:
        #     request.headers.update(self.settings.custom_headers)
        yield request

    async def async_auth_flow(self, request: aiohttpx.Request):
        """
        Injects the API Key into the Request
        """
        if self.auth_key not in request.headers:
            request.headers[self.auth_key] = self.auth_value
        if custom_headers := self.context.custom_headers:
            request.headers.update(custom_headers)
        # request.headers.update(self.settings.get_api_key_headers())
        # request.headers[self.auth_key] = self.auth_value
        # if self.settings.custom_headers:
        #     request.headers.update(self.settings.custom_headers)
        yield request


class ExternalProviderAuth(aiohttpx.Auth):
    """
    Custom Authentication Wrapper for External OpenAI Clients
    """
    def __init__(
        self, 
        config: 'ExternalProviderConfig',
        is_proxied: Optional[bool] = None,
        **kwargs,
    ):
        """
        Initializes the External Provider Auth

        :TODO - add support for Proxy
        """
        self.config = config
        self.is_proxied = is_proxied

    def get_api_key(self) -> Optional[str]:
        """
        Returns the API Key
        """
        if self.config.has_api_keys:
            if len(self.config.api_keys) == 1:
                return self.config.api_keys_value[0]
            return random.choice(self.config.api_keys_value)
        return self.config.api_key_value
    
    @property
    def has_api_key(self) -> bool:
        """
        Returns whether the API Key is present
        """
        return self.config.has_api_key or self.config.has_api_keys

    def auth_flow(self, request: aiohttpx.Request):
        """
        Injects the API Key into the Request
        """
        if self.has_api_key and self.config.api_key_header not in request.headers:
            request.headers[self.config.api_key_header] = self.get_api_key()
        if self.config.custom_headers:
            request.headers.update(self.config.custom_headers)
        if self.is_proxied and self.config.proxy_headers:
            request.headers.update(self.config.proxy_headers)
        yield request

    async def async_auth_flow(self, request: aiohttpx.Request):
        """
        Injects the API Key into the Request
        """
        if self.has_api_key and self.config.api_key_header not in request.headers:
            request.headers[self.config.api_key_header] = self.get_api_key()
        if self.config.custom_headers:
            request.headers.update(self.config.custom_headers)
        if self.is_proxied and self.config.proxy_headers:
            request.headers.update(self.config.proxy_headers)
        yield request
        
        


from __future__ import annotations

"""
Reranking Router
"""

from typing import Optional, Type, Any, Union, List, Dict, overload
from lzo.types import (
    eproperty, 
    validator, 
    root_validator, 
    field_validator,
    Field,
    PYDANTIC_VERSION, 
)

from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.types.base import BaseResource
from lzl.api.openai.types.responses import BaseResponse
from lzl.api.openai.types.routes import BaseRoute
from lzl.api.openai.utils import logger

__all__ = [
    'RerankingData',
    'RerankingObject',
    'RerankingResponse',
    'RerankingRoute',
]

# Currently only voyager supports this.

class RerankingData(BaseResource):
    relevance_score: Optional[float] = None
    document: Optional[str] = None
    index: Optional[int] = 0

class RerankingObject(BaseResource):
    model: Optional[str] = None
    query: Optional[str] = None
    documents: Optional[List[Any]] = None
    top_k: Optional[int] = None
    # return_documents: Optional[bool] = Field(False, description = 'If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.') 
    # truncation: Optional[bool] = Field(True, description = 'Whether to truncate the input text to the maximum context length')

    @field_validator('model', mode = 'before')
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'): v = values.get('engine')
            elif values.get('deployment'): v = values.get('deployment')
        
        model = ModelContextHandler.get(v)
        if not model.data.get('supports_reranking') and 'rerankings' not in model.endpoints:
            raise ValueError(f"Model {model.name} does not support reranking")
        return model.name


    def dict(self, *args, exclude: Any = None, exclude_unset: bool = True, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, exclude_unset = exclude_unset, **kwargs)
    


class RerankingResponse(BaseResponse):
    data: Optional[List[RerankingData]] = None
    data_model: Optional[Type[BaseResource]] = RerankingData
    input_object: Optional[RerankingObject] = None

    @eproperty
    def openai_model(self):
        """
        Returns the model for the completions
        """
        return self.headers.get('openai-model', self.input_object.model)

    @eproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """ 
        return ModelContextHandler.get_consumption_cost(
            model_name = self.openai_model,
            usage = self.usage,
        )



class RerankingRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = RerankingObject
    response_model: Optional[Type[BaseResource]] = RerankingResponse
    api_resource: Optional[str] = Field(default = 'rerank')
    root_name: Optional[str] = Field(default = 'rerank')

    @eproperty
    def create_enabled(self):
        """
        Returns whether the Create Route is Enabled
        """
        # This is disabled by default
        return False

    @overload
    def create(
        self, 
        query: Optional[str] = None,
        documents: Optional[List[Any]] = None,
        top_k: Optional[int] = None,
        return_documents: Optional[bool] = False,
        truncation: Optional[bool] = True,
        model: Optional[str] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> RerankingResponse:
        """
        Creates a reranking response for the provided query and documents

        Usage:

        ```python
        >>> result = OpenAI.reranking.create(
        >>>    query = 'say this is a test',
        >>>    documents = ['say this is a test', 'say this is another test'],
        >>> )
        ```

        **Parameters:**

        :query (string, required): The query to rerank
        :documents (list[string], required): The documents to rerank
        :top_k (int, optional): The number of documents to return
        :return_documents (bool, optional): If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.
        :truncation (bool, optional): Whether to truncate the input text to the maximum context length
        :model (string, optional): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.

        Returns: `RerankingResponse`
        """
        ...


    def create(
        self, 
        input_object: Optional[RerankingObject] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> RerankingResponse:
        """
        Creates a reranking response for the provided query and documents

        Usage:

        ```python
        >>> result = OpenAI.reranking.create(
        >>>    query = 'say this is a test',
        >>>    documents = ['say this is a test', 'say this is another test'],
        >>> )
        ```

        **Parameters:**

        :query (string, required): The query to rerank
        
        :documents (list[string], required): The documents to rerank
        
        :top_k (int, optional): The number of documents to return
        
        :return_documents (bool, optional): If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.
        
        :truncation (bool, optional): Whether to truncate the input text to the maximum context length
        
        :model (string, optional): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.

        Returns: `RerankingResponse`
        """
        return super().create(
            input_object = input_object,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            **kwargs
        )

    @overload
    async def async_create(
        self, 
        query: Optional[str] = None,
        documents: Optional[List[Any]] = None,
        top_k: Optional[int] = None,
        return_documents: Optional[bool] = False,
        truncation: Optional[bool] = True,
        model: Optional[str] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> RerankingResponse:
        """
        Creates a reranking response for the provided query and documents

        Usage:

        ```python
        >>> result = await OpenAI.reranking.async_create(
        >>>    query = 'say this is a test',
        >>>    documents = ['say this is a test', 'say this is another test'],
        >>> )
        ```

        **Parameters:**

        :query (string, required): The query to rerank
        
        :documents (list[string], required): The documents to rerank
        
        :top_k (int, optional): The number of documents to return
        
        :return_documents (bool, optional): If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.
        
        :truncation (bool, optional): Whether to truncate the input text to the maximum context length
        
        :model (string, optional): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.

        Returns: `RerankingResponse`
        """
        ...


    async def async_create(
        self, 
        input_object: Optional[RerankingObject] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs
    ) -> RerankingResponse:  # sourcery skip: low-code-quality
        """
        Creates a reranking response for the provided query and documents

        Usage:

        ```python
        >>> result = await OpenAI.reranking.async_create(
        >>>    query = 'say this is a test',
        >>>    documents = ['say this is a test', 'say this is another test'],
        >>> )
        ```

        **Parameters:**

        :query (string, required): The query to rerank
        
        :documents (list[string], required): The documents to rerank
        
        :top_k (int, optional): The number of documents to return
        
        :return_documents (bool, optional): If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.
        
        :truncation (bool, optional): Whether to truncate the input text to the maximum context length
        
        :model (string, optional): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.

        Returns: `RerankingResponse`
        """
        return await super().async_create(
            input_object = input_object,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            **kwargs
        )
        



from __future__ import annotations

"""
Common Environment Variables
"""

import os
from pathlib import Path

GITHUB_TOKEN: str = os.environ.get('GH_TOKEN', os.getenv('GITHUB_TOKEN', ''))

# Paths
TEMP_PATH = os.getenv('TEMP_PATH', '/tmp')
BUILD_CONFIG_PATH = Path(os.getenv('BUILD_CONFIG_PATH', f'{TEMP_PATH}/build_config.yaml'))

LIB_PATH = Path(__file__).parent
ASSETS_PATH = LIB_PATH.joinpath('assets')

APP_HOME = os.getenv('APP_HOME', '/app')
APP_PATH = Path(APP_HOME)

REQUIREMENTS_PATH = Path(os.getenv('REQUIREMENTS_PATH', '/tmp/requirements'))

if pkgs_path := os.getenv('PKGS_PATH'):
    PKGS_PATH = Path(pkgs_path)
else:
    PKGS_PATH = REQUIREMENTS_PATH.joinpath('pkgs')

# DEFAULT_CONFIG_PATH = _HERE.joinpath('default_build_config.yaml')



from __future__ import annotations

"""Client registry utilities for the LazyOps orchestration layer."""

import typing as t

from .base import MRegistry, combine_parts

if t.TYPE_CHECKING:
    from .mixins import RegisteredClient

    RClientT = t.TypeVar('RClientT', bound='RegisteredClient')
else:
    RClientT = t.TypeVar('RClientT')

__all__ = [
    'get_app_client',
    'register_app_client',
    'register_app_clients',
    'register_client',
    'register_client_hook',
    'register_initialized_client',
]

_cregistry: MRegistry['RClientT'] = MRegistry('clients')


def register_initialized_client(client: 'RClientT', **kwargs: t.Any) -> None:
    """Record an already instantiated client in the registry.

    Args:
        client: Concrete instance created by the caller.
        **kwargs: Currently unused; retained for API compatibility.
    """

    if hasattr(client, '_rxtra'):
        _cregistry._register_initialized(client._rxtra['registry_name'], client)
        return
    cls_name = _cregistry.get_classname(client, is_type=False)
    cls_module = (
        client._rmodule
        if getattr(client, '_rmodule', None) is not None
        else client.__class__.__module__.split('.')[0]
    )
    cls_submodule = (
        client._rsubmodule if getattr(client, '_rsubmodule', None) is not None else None
    )
    client_name = client.name if getattr(client, 'name', None) is not None else cls_name

    registry_name = combine_parts(cls_module, cls_submodule, client_name)
    if registry_name in _cregistry.init_registry:
        return
    _cregistry._register_initialized(registry_name, client)


def register_client(client: t.Union[t.Type['RClientT'], 'RClientT'], **kwargs: t.Any) -> None:
    """Register a client class or instance for lazy lookup.

    Args:
        client: Client type or pre-instantiated object to register.
        **kwargs: Ignored but accepted for compatibility.
    """

    if not isinstance(client, type):
        register_initialized_client(client, **kwargs)
        return
    cls_name = _cregistry.get_classname(client, is_type=True)
    client_name = client.name if getattr(client, 'name', None) is not None else cls_name
    cls_module = (
        client._rmodule
        if getattr(client, '_rmodule', None) is not None
        else client.__module__.split('.')[0]
    )
    cls_submodule = client._rsubmodule if getattr(client, '_rsubmodule', None) is not None else None
    registry_name = combine_parts(cls_module, cls_submodule, client_name)
    if registry_name in _cregistry.mregistry:
        _cregistry.logger.warning(
            f'Client {client_name} already registered with `{registry_name}`'
        )
        return
    client._rxtra['module'] = cls_module
    client._rxtra['submodule'] = cls_submodule
    client._rxtra['cls_name'] = cls_name
    client._rxtra['client_name'] = client_name
    client._rxtra['registry_name'] = registry_name
    client._rxtra['module_path'] = _cregistry.get_module_path(client)
    client._rxtra['registered'] = True
    _cregistry[registry_name] = client


def register_app_client(
    client_name: str,
    client_path: str,
    module: t.Optional[str] = None,
    submodule: t.Optional[str] = None,
) -> None:
    """Register a lazily importable client path.

    Args:
        client_name: Friendly name used for lookup.
        client_path: Dotted import path to the client class.
        module: Optional module prefix used to namespacing entries.
        submodule: Additional namespace segment for nested registries.
    """

    registry_name = combine_parts(module, submodule, client_name)
    if registry_name in _cregistry.mregistry:
        return
    _cregistry.uninit_registry[registry_name] = client_path


def register_app_clients(
    clients: t.Mapping[str, str],
    module: t.Optional[str] = None,
    submodule: t.Optional[str] = None,
) -> None:
    """Bulk register lazily loaded client paths.

    Args:
        clients: Mapping of client names to dotted import paths.
        module: Optional module prefix shared across entries.
        submodule: Optional nested namespace for the group.
    """

    for client_name, client_path in clients.items():
        register_app_client(client_name, client_path, module=module, submodule=submodule)


def get_app_client(
    client_name: str,
    module: t.Optional[str] = None,
    submodule: t.Optional[str] = None,
    **kwargs: t.Any,
) -> 'RegisteredClient':
    """Retrieve a client by name, instantiating it if needed.

    Args:
        client_name: Identifier passed to :func:`register_client`.
        module: Optional module prefix used during registration.
        submodule: Optional nested namespace used during registration.
        **kwargs: Forwarded to the registered constructor when lazily importing.

    Returns:
        Registered client instance associated with ``client_name``.
    """

    registry_name = combine_parts(module, submodule, client_name)
    return _cregistry.get(registry_name, **kwargs)


def register_client_hook(
    client_name: str,
    func: t.Union[t.Callable[..., t.Any], str],
    kind: t.Literal['pre', 'post'] = 'post',
    module: t.Optional[str] = None,
    submodule: t.Optional[str] = None,
) -> None:
    """Attach a hook to a registered client.

    Args:
        client_name: Registry key that identifies the client.
        func: Callable or dotted import path executed as the hook.
        kind: Whether to run the hook before (``'pre'``) or after (``'post'``)
            instantiation.
        module: Optional module prefix used during registration.
        submodule: Optional nested namespace used during registration.
    """

    registry_name = combine_parts(module, submodule, client_name)
    _cregistry.register_hook(registry_name, func, kind=kind)



from __future__ import annotations

"""Helpers for managing IO-related registries.

This module centralises the lazily-evaluated serializers and the lifecycle of
temporary files.  Documenting the behaviour here helps keep downstream tooling
aware of the implicit global state that exists around serializer lookups and
file clean-up hooks.
"""

import contextlib
import typing as t

from lzl.proxied import ProxyObject

if t.TYPE_CHECKING:
    from lzl.io import FileLike
    from lzl.io.ser import JsonSerializer

__all__ = [
    "get_json_ser",
    "get_jsonb_ser",
    "json_ser",
    "jsonb_ser",
    "add_file_to_cleanup",
    "cleanup_files_on_exit",
    "acleanup_files_on_exit",
]


def get_json_ser(*args: t.Any, **kwargs: t.Any) -> "JsonSerializer":
    """Return the canonical JSON serializer instance.

    The function intentionally accepts arbitrary positional and keyword
    arguments so that it can be used as a drop-in replacement for serializer
    factories throughout the codebase.  Any provided values are ignored because
    the serializer configuration is controlled globally via LazyOps settings.
    """

    from lzl.io.ser import get_serializer

    return get_serializer("json")


def get_jsonb_ser(*args: t.Any, **kwargs: t.Any) -> "JsonSerializer":
    """Return a JSON serializer that emits byte values (`jsonb`).

    Similar to :func:`get_json_ser`, the arguments are accepted purely for API
    compatibility.  The serializer is configured to enforce byte output so it
    can be used with backends expecting binary payloads.
    """

    from lzl.io.ser import get_serializer

    return get_serializer("json", enforce_byte_value=True)


jsonb_ser: "JsonSerializer" = ProxyObject(obj_getter=get_jsonb_ser)
json_ser: "JsonSerializer" = ProxyObject(obj_getter=get_json_ser)

_registered_file_aexit: bool = False
_registered_file_exit: bool = False

_created_tempfiles: t.Set["FileLike"] = set()


def cleanup_files_on_exit() -> None:
    """Remove temporary files that were created through LazyOps IO helpers."""

    while _created_tempfiles:
        file_path = _created_tempfiles.pop()
        with contextlib.suppress(Exception):
            file_path.unlink(missing_ok=True)


async def acleanup_files_on_exit() -> None:
    """Asynchronous variant of :func:`cleanup_files_on_exit`."""

    while _created_tempfiles:
        file_path = _created_tempfiles.pop()
        with contextlib.suppress(Exception):
            await file_path.aunlink(missing_ok=True)


def register_file_exit(*_args: t.Any, **_kwargs: t.Any) -> None:
    """Register the clean-up hooks exactly once for the process lifecycle."""

    global _registered_file_exit, _registered_file_aexit
    if not _registered_file_exit:
        import atexit

        atexit.register(cleanup_files_on_exit)
        _registered_file_exit = True
    # TODO: add async clean-up hook when aioexit dependency is restored.
    _registered_file_aexit = _registered_file_aexit or False


def add_file_to_cleanup(file: "FileLike") -> None:
    """Track a file so it is safely removed when the interpreter exits."""

    register_file_exit()
    _created_tempfiles.add(file)



from __future__ import annotations

"""
Standalone Logging Module that doesn't depend on any other modules
"""

import re
import os
import sys
import abc
import logging
import typing
import traceback
import warnings
import pprint
import atexit as _atexit
import functools
import threading
import contextlib
from enum import Enum

from loguru import _defaults
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger
from .static import DEFAULT_STATUS_COLORS, QUEUE_STATUS_COLORS, STATUS_COLOR, FALLBACK_STATUS_COLOR, DEFAULT_FUNCTION_COLOR, DEFAULT_CLASS_COLOR, RESET_COLOR, LOGLEVEL_MAPPING, REVERSE_LOGLEVEL_MAPPING, COLORED_MESSAGE_MAP
from .utils import format_item, format_message, get_logging_level
from .state import is_registered_logger_module, is_global_muted
from .formatters import LoggerFormatter
from .mixins import LoggingMixin
from typing import Type, Union, Optional, Any, List, Dict, Tuple, Callable, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from .utils import MsgItem

# Use this section to filter out warnings from other modules
os.environ['TF_CPP_MIN_LOG_LEVEL'] = os.getenv('TF_CPP_MIN_LOG_LEVEL', '3')
warnings.filterwarnings('ignore', message='Unverified HTTPS request')
warnings.filterwarnings("ignore", message = "Unclosed client session")

if TYPE_CHECKING:
    from pydantic_settings import BaseSettings


# Setup Default Logger
class Logger(_Logger, LoggingMixin):

    name: str = None
    settings: Type['BaseSettings'] = None
    conditions: Dict[str, Tuple[Union[Callable, bool], str]] = {}
    default_trace_depth: Optional[int] = None
    is_global: bool = False
    _colored_opts = None
    _current_level: Optional[str] = None

    @property
    def colored_opts(self):
        """
        Returns the colored options
        """
        if not self._colored_opts:
            (exception, depth, record, lazy, colors, raw, capture, patchers, extra) = self._options
            self._colored_opts = (exception, depth, record, lazy, True, raw, capture, patchers, extra)
        return self._colored_opts

    def _get_opts(self, colored: Optional[bool] = False, **kwargs):
        """
        Returns the options
        """
        return self.colored_opts if colored else self._options

    def get_log_mode(self, level: str = "info"):
        """
        Returns the log mode based on the level
        """
        return self.dev if level.upper() in {'DEV'} else getattr(self, level.lower())

    def add_if_condition(
        self, 
        name: str, 
        condition: Union[Callable, bool],
        level: Optional[Union[str, int]] = 'INFO',
    ):
        """
        Adds a condition to the logger
        """
        self.conditions[name] = (condition, self._get_level(level))
    
    def remove_if_condition(self, name: str):
        """
        Removes a condition from the logger
        """
        if name in self.conditions:
            del self.conditions[name]

    def _is_dev_condition(self, record: logging.LogRecord) -> bool:
        """
        Returns whether the dev condition is met
        """
        if not self.settings: return True
        if record.levelname == 'DEV':
            for key in {'api_dev_mode', 'debug_enabled'}:
                if (
                    hasattr(self.settings, key)
                    and getattr(self.settings, key) is False
                ):
                    return False
        return True

    def _filter_if(self, name: str, record: Optional[logging.LogRecord] = None, message: Optional[Any] = None, level: Optional[Union[str, int]] = None) -> Tuple[bool, str]:
        """
        Filters out messages based on conditions
        """
        if name in self.conditions:
            condition, clevel = self.conditions[name]
            if isinstance(condition, bool):
                return condition, clevel
            elif isinstance(condition, type(None)):
                return False, clevel
            elif isinstance(condition, Callable):
                return condition(record or message), clevel
        return True, (record.levelname if record else self._get_level(level or 'INFO'))
    
    def _filter_module_name(self, name: str) -> bool:
        """
        Filter based on module name

        - True if the module is not registered and is_global is False 
        - False if the module is registered and is_global is False
        """
        _is_registered = is_registered_logger_module(name)
        if self.is_global: 
            return _is_registered is not False
        return _is_registered is False


    def _filter(self, record: logging.LogRecord, name: Optional[str] = None) -> bool:
        """
        Filters out messages based on conditions

        - True if the message should be filtered out
        - False if the message should be logged
        """
        if self.check_silenced(record):
            return True
        if self._filter_module_name(record['name']): 
            return True
        
        if name is not None:
            return self._filter_if(name, record)[0]
        if not self.conditions: return False
        return not any(
            isinstance(value, bool)
            and value is False
            or not isinstance(value, bool)
            and isinstance(value, Callable)
            and value(record) is False
            for key, value in self.conditions.items()
        )

    def _filter_dev(self, record: logging.LogRecord, **kwargs):
        if not self.settings:
            return True
        if record.levelname == 'DEV':
            for key in {'api_dev_mode', 'debug_enabled'}:
                if (
                    hasattr(self.settings, key)
                    and getattr(self.settings, key) is False
                ):
                    return False
        return True
    
    def opt(
        self,
        *,
        exception=None,
        record=False,
        lazy=False,
        colors=False,
        raw=False,
        capture=True,
        depth=0,
        ansi=False
    ):
        """
        Return a new logger with the specified options changed.
        """
        if ansi: colors = True
        args = self._options[-2:]
        return type(self)(self._core, exception, depth, record, lazy, colors, raw, capture, *args)



    """
    Newly Added APIs
    """

    def _get_level(self, level: Union[str, int]) -> str:
        """
        Returns the log level
        """
        return get_logging_level(level)

    def _format_item(
        self,
        msg: 'MsgItem',
        max_length: Optional[int] = None,
        colored: Optional[bool] = False,
        level: Optional[str] = None,
        _is_part: Optional[bool] = False,
    ) -> str:  # sourcery skip: extract-duplicate-method, low-code-quality, split-or-ifs
        """
        Formats an item
        """
        return format_item(msg, max_length = max_length, colored = colored, level = level, _is_part = _is_part)


    def _format_message(
        self, 
        message: 'MsgItem',
        *args,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        level: Optional[str] = None,
        colored: Optional[bool] = False,
        extra: Optional[Dict[str, Any]] = None,
    ) -> str:
        """
        Formats the message

        "example |b|msg|e|"
        -> "example <blue>msg</><reset>"
        """
        return format_message(
            message,
            *args,
            prefix = prefix,
            max_length = max_length,
            level = level,
            colored = colored,
            extra = extra,
        )
    
    def log_if(
        self, 
        name: str, 
        message: 'MsgItem',
        *args, 
        level: Optional[Union[str, int]] = None, 
        **kwargs
    ):  # noqa: N805
        """
        Log ``message.format(*args, **kwargs)`` with severity ``level`` if condition is met.
        """
        condition, clevel = self._filter_if(name, message = message, level = level)
        if condition:
            return self.log((level or clevel), message, *args, **kwargs)

    def log(
        self, 
        level: Union[str, int], 
        message: 'MsgItem',
        *args, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        colored: Optional[bool] = False,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        **kwargs
    ):  # noqa: N805
        """
        Log ``message.format(*args, **kwargs)`` with severity ``level``.
        """
        level = self._get_level(level)
        extra = kwargs.pop('extra', None)
        message = self._format_message(
            message,
            prefix = prefix,
            max_length = max_length,
            colored = colored,
            level = level,
            extra = extra,
        )
        try:
            self._log(level, False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            # level_id, static_level_no, from_decorator, options, message, args, kwargs
            static_log_no = REVERSE_LOGLEVEL_MAPPING.get(level, 20)
            self._log(level, static_log_no, False, self._get_opts(colored = colored), message, args, kwargs)
        self.run_logging_hooks(message, hook = hook)

    def info(
        self, 
        message: 'MsgItem',
        *args, 
        colored: Optional[bool] = None, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        **kwargs
    ):  # noqa: N805
        """
        Log ``message.format(*args, **kwargs)`` with severity ``'INFO'``.
        """
        if colored is None and isinstance(message, str) and '|e|' in message: colored = True
        extra = kwargs.pop('extra', None)
        message = self._format_message(
            message,
            *args,
            prefix = prefix,
            max_length = max_length,
            colored = colored,
            level = 'INFO',
            extra = extra,
        )
        if not is_global_muted():
            try:
                self._log("INFO", False, self._get_opts(colored = colored), message, args, kwargs)
            except TypeError:
                # Compatibility with < 0.6.0
                self._log("INFO", 20, False, self._get_opts(colored = colored), message, args, kwargs)
        self.run_logging_hooks(message, hook = hook)

    def success(
        self, 
        message: 'MsgItem', 
        *args, 
        colored: Optional[bool] = False, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        **kwargs
    ):  # noqa: N805
        r"""Log ``message.format(*args, **kwargs)`` with severity ``'SUCCESS'``."""
        extra = kwargs.pop('extra', None)
        message = self._format_message(
            message,
            *args,
            prefix = prefix,
            max_length = max_length,
            colored = colored,
            level = 'SUCCESS',
            extra = extra,
        )
        if not is_global_muted():
            try:
                self._log("SUCCESS", False, self._get_opts(colored = colored), message, args, kwargs)
            except TypeError:
                # Compatibility with < 0.6.0
                self._log("SUCCESS", 20, False, self._get_opts(colored = colored), message, args, kwargs)
        self.run_logging_hooks(message, hook = hook)

    def warning(
        self, 
        message, 
        *args, 
        colored: Optional[bool] = False, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        **kwargs
    ):  # noqa: N805
        r"""Log ``message.format(*args, **kwargs)`` with severity ``'WARNING'``."""
        extra = kwargs.pop('extra', None)
        message = self._format_message(
            message,
            prefix = prefix,
            max_length = max_length,
            colored = colored,
            level = 'WARNING',
            extra = extra,
        )
    
        try:
            self._log("WARNING", False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            self._log("WARNING", 30, False, self._get_opts(colored = colored), message, args, kwargs)
        self.run_logging_hooks(message, hook = hook)

    def error(
        self,
        message: Any,
        *args,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        colored: Optional[bool] = False,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        exc_info: Optional[bool] = False,
        **kwargs
    ) -> None:
        """
        Log ``message.format(*args, **kwargs)`` with severity ``'ERROR'``.
        """
        extra = kwargs.pop('extra', None)
        message = self._format_message(
            message,
            prefix = prefix,
            max_length = max_length,
            colored = colored,
            level = 'ERROR',
            extra = extra,
        )
        if exc_info: message += f"\n{traceback.format_exc()}"

        try:
            self._log("ERROR", False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            self._log("ERROR", 40, False, self._get_opts(colored = colored), message, args, kwargs)
        self.run_logging_hooks(message, hook = hook)

    def trace(
        self, 
        msg: 'MsgItem',
        error: Optional[Type[Exception]] = None, 
        level: str = "ERROR",
        limit: Optional[int] = None,
        chain: Optional[bool] = True,
        colored: Optional[bool] = False,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        **kwargs,
    ) -> None:
        """
        This method logs the traceback of an exception.

        :param error: The exception to log.
        """
        _depth = kwargs.pop('depth', None)
        extra = kwargs.pop('extra', None)
        if _depth is not None: limit = _depth
        if isinstance(msg, str):
            _msg = msg
            if extra:
                extras_rendered = format_item(extra, max_length = max_length, colored = colored, level = level)
                extras_rendered = extras_rendered.lstrip('\n')
                if extras_rendered:
                    if _msg:
                        _msg += f"\n{extras_rendered}"
                    else:
                        _msg = extras_rendered
        else:
            _msg = self._format_message(
                msg,
                colored = colored,
                level = level,
                prefix = prefix,
                max_length = max_length,
                extra = extra,
            )
        # pprint.pformat(msg)
        _msg += f"\n{traceback.format_exc(chain = chain, limit = limit)}"
        if error: _msg += f" - {error}"
        
        try:
            self._log(level, False, self._get_opts(colored = colored), _msg, (), {})
        except TypeError:
            static_log_no = REVERSE_LOGLEVEL_MAPPING.get(level, 40)
            self._log(level, static_log_no, False, self._get_opts(colored = colored), _msg, (), {})
        self.run_logging_hooks(_msg, hook = hook)

    def exception(
        self,
        message: 'MsgItem',
        *args,
        colored: Optional[bool] = False,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        hook: Optional[Union[Callable, List[Callable]]] = None,
        **kwargs
    ):
        """
        Log ``message.format(*args, **kwargs)`` with severity ``'ERROR'``.
        """
        extra = kwargs.get('extra')
        message = self._format_message(
            message,
            *args,
            prefix = prefix,
            max_length = max_length,
            colored = colored,
            level = 'ERROR',
            extra = extra,
        )
        super().exception(message, *args, **kwargs)
        self.run_logging_hooks(message, hook = hook)

    
    def __call__(self, message: 'MsgItem', *args, level: str = 'info', **kwargs):
        r"""Log ``message.format(*args, **kwargs)`` with severity ``'INFO'``."""
        if isinstance(message, list):
            __message = "".join(f'- {item}\n' for item in message)
        elif isinstance(message, dict):
            __message = "".join(f'- {key}: {value}\n' for key, value in message.items())
        else:
            __message = str(message)
        _log = self.get_log_mode(level)
        _log(__message.strip(), *args, **kwargs)
    
    def _logcompat(
        self, level, from_decorator, options, message, args, kwargs
    ):
        """
        Compatible to < 0.6.0
        """
        try:
            self._log(level, from_decorator, options, message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            # level_id, static_level_no, from_decorator, options, message, args, kwargs
            static_log_no = REVERSE_LOGLEVEL_MAPPING.get(level, 20)
            self._log(level, static_log_no, from_decorator, options, message, args, kwargs)


    """
    Utilties
    """

    def change_logger_level(
        self,
        level: str,
    ):
        """
        Changes the logger level

        :TODO
        """
        return
        # self._core.configure(levels = [level])

    

class InterceptHandler(logging.Handler):
    loglevel_mapping = LOGLEVEL_MAPPING

    def emit(self, record):
        try:
            from .main import logger
        except ImportError:
            return
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = self.loglevel_mapping.get(record.levelno, 'DEBUG')
        # if "Unclosed client session" in record.message:
        #     print('Has unclosed client session')
        #     return
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        log = logger.bind(request_id=None)
        log.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())



from __future__ import annotations

import abc
import logging
import warnings
import pprint
import contextlib
from .state import register_module_name
from typing import Optional, Set, Callable, TYPE_CHECKING, Union, List, Type, Tuple, Any, Dict

if TYPE_CHECKING:
    from .base import Logger


def temp_silence_filter(record: logging.LogRecord) -> int:
    """
    Temporary silence filter
    """
    return 0

class LoggingMixin(abc.ABC):
    """
    A mixin class for logging hooks
    """

    _logging_hooks: Optional[Set[Callable]] = set()
    _silenced_modules: Optional[Set[str]] = set()
    _temp_silenced_modules: Dict[str, str] = {}
    _temp_silenced_loggers: Dict[str, Tuple[logging.Logger, int]] = {}

    if TYPE_CHECKING:
        def get_log_mode(self, level: Union[str, int]) -> str:
            ...

    def add_logging_hook(self, *hooks: Callable):
        """
        Adds a logging hook
        """
        for hook in hooks:
            self._logging_hooks.add(hook)

    def remove_logging_hook(self, *hooks: Callable):
        """
        Removes a logging hook
        """
        for hook in hooks:
            self._logging_hooks.remove(hook)

    def add_silenced_modules(self, *modules: str):
        """
        Adds a list of modules to the silenced modules
        """
        for module in modules:
            self._silenced_modules.add(module)

    def remove_silenced_modules(self, *modules: str):
        """
        Removes a list of modules from the silenced modules
        """
        for module in modules:
            self._silenced_modules.remove(module)
            self.remove_temp_silence_from_logging_module(module)
    
    def run_logging_hooks(self, message: str, hook: Optional[Union[Callable, List[Callable]]] = None):
        """
        Runs the logging hooks
        """
        for log_hook in self._logging_hooks:
            log_hook(message)
        if hook: 
            if not isinstance(hook, list): hook = [hook]
            for h in hook:
                h(message)

    @contextlib.contextmanager
    def hooks(self, *hooks: Callable):
        """
        Adds a logging hook
        """
        self.add_logging_hook(*hooks)
        try:
            yield
        finally:
            self.remove_logging_hook(*hooks)

    @contextlib.contextmanager
    def silenced(self, *modules: str):
        """
        Adds a list of modules to the silenced modules
        """
        self.add_silenced_modules(*modules)
        try:
            yield
        finally:
            self.remove_silenced_modules(*modules)

    def set_module_name(self, src_name: str, module_name: str, is_relative: bool = False):
        """
        Sets the module name
        """
        register_module_name(src_name, module_name, is_relative = is_relative)

    """
    This part still doesn't work atm.
    """
    def is_silenced_record(self, record: Dict[str, Any]) -> Tuple[str, bool]:
        """
        Checks if the record is silenced
        """
        if '.' in record['name']:
            src_module = record['name'].split('.')[0]
            return src_module, src_module in self._silenced_modules
        return record['name'], record['name'] in self._silenced_modules

    def add_temp_silence_to_logging_module(self, src_module: str, module: str):
        """
        Adds a module to the temp silenced modules
        """
        _logger = logging.getLogger(src_module)
        # _logger.addFilter(temp_silence_filter)
        self._temp_silenced_loggers[src_module] = (_logger, _logger.level)
        _logger.setLevel(logging.ERROR)

    def remove_temp_silence_from_logging_module(self, src_module: str):
        """
        Removes a module from the temp silenced modules
        """
        # module = self._temp_silenced_modules.pop(src_module, None)
        (_logger, level) = self._temp_silenced_loggers.pop(src_module, (None, None))
        if not _logger: return
        _logger.setLevel(level)
        # if not module: return
        # _logger = logging.getLogger(module)
        # _logger.removeFilter(temp_silence_filter)

    def check_silenced(self, record: Dict[str, Any]) -> bool:
        """
        Checks if the record is silenced
        """
        src_module, is_silenced = self.is_silenced_record(record)
        if is_silenced and src_module not in self._temp_silenced_loggers:
            self.add_temp_silence_to_logging_module(src_module, record['name'])
        return is_silenced


    """
    Rendering Utilities
    """
    def render(
        self,
        objs: Union[Any, List[Any]],
        level: str = 'INFO',
        **kwargs,
    ) -> None:
        """
        Tries to render an object with pformat
        """
        if not isinstance(objs, list): objs = [objs]
        _log = self.get_log_mode(level)
        for obj in objs:
            try: _log('\n' + pprint.pformat(obj, **kwargs))
            except Exception as e: _log('\n' + obj)

    def render_yaml(
        self,
        objs: Union[Any, List[Any]],
        level: str = 'INFO',
        **kwargs,
    ) -> None:
        """
        Tries to render a yaml object.
        """
        _log = self.get_log_mode(level)
        try:
            from lzl.io.ser import Yaml
            _log('\n' + Yaml.dumps(objs, **kwargs))
        
        except Exception as e:
            import yaml
            _log('\n' + yaml.dump(objs, **kwargs))


    def display_crd(self, message: Any, *args, level: str = 'info', **kwargs):
        """
        Display CRD information in the log.

        Parameters
        ----------
        message : Any
            The message to display.
        level : str
            The log level to use.
        """
        __message = ""
        if isinstance(message, list):
            for m in message:
                if isinstance(m, dict):
                    __message += "".join(f'- <light-blue>{key}</>: {value}\n' for key, value in m.items())
                else:
                    __message += f'- <light-blue>{m}</>\n'
        elif isinstance(message, dict):
            __message = "".join(f'- <light-blue>{key}</>: {value}\n' for key, value in message.items())
        else:
            __message = str(message)
        _log = self.get_log_mode(level)
        _log(__message.strip(), *args, **kwargs)



    """
    Logging Utilities
    """

    def mute_logger(self, modules: Optional[Union[str, List[str]]], level: str = 'WARNING'):
        """
        Helper to mute a logger from another module.
        """
        if not isinstance(modules, list): modules = [modules]
        for module in modules:
            logging.getLogger(module).setLevel(logging.getLevelName(level))

    def mute_warning(self, action: str = 'ignore', category: Type[Warning] = Warning, module: str = None, **kwargs):
        """
        Helper to mute a warning from another module.
        """
        if module: warnings.filterwarnings(action, category=category, module=module, **kwargs)
        else: warnings.filterwarnings(action, category=category, **kwargs)
    
    def add_healthz_filter(
        self,
        loggers: Optional[Union[str, List[str]]] = None,
        paths: Optional[Union[str, List[str]]] = None,
    ):
        """
        Adds a filter to the logger to filter out healthz requests.

        Parameters
        ----------
        loggers : Optional[Union[str, List[str]]]
            The loggers to add the filter to.
            defaults to ['gunicorn.glogging.Logger', 'uvicorn.access']
        paths : Optional[Union[str, List[str]]]
            The paths to filter out.
            defaults to ['/healthz', '/health']
        """
        if not loggers: loggers = ['gunicorn.glogging.Logger', 'uvicorn.access']
        if not paths: paths = ['/healthz', '/health']
        if not isinstance(loggers, list): loggers = [loggers]
        if not isinstance(paths, list): paths = [paths]

        def _healthz_filter(record: logging.LogRecord) -> bool:
            if 'path' in record.args:
                # if record.args.get('path'):
                return record.args['path'] not in paths
            return all(path not in record.args for path in paths)

        for logger in loggers:
            logging.getLogger(logger).addFilter(_healthz_filter)

    def mute_module(
        self,
        module: str,
        **kwargs,
    ):
        """
        Efectively mutes a module
            
            - `module`: The module to mute
        """
        def _mute_filter(record: logging.LogRecord) -> bool:
            return record.module == module
        logging.getLogger(module).addFilter(_mute_filter)
    


from __future__ import annotations

"""
Base Workflow Context Object
"""

import abc
import tracemalloc
import contextlib
from .utils import capture_state, display_top
from typing import Any, Dict, List, Optional, TypeVar, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from kvdb import PersistentDict, AsyncLock, KVDBSession
    from lzo.utils.logs import Logger
    from lzl.api.hatchet.session import HatchetSession
    from lzl.api.hatchet.config import HatchetSettings
    from lzl.api.hatchet.context import Context
    from lazyops.libs.abcs.types.persistence import TemporaryData
    from ..base import WorkflowT
    


class BaseWorkflowContext(abc.ABC):
    """
    Base class for all Workflow Context
    """
    _extra: Dict[str, Any] = {}
    _clients: Optional[Dict[str, Any]] = {}
    _pdicts: Dict[str, 'PersistentDict'] = {}

    def __init__(
        self, 
        workflow: 'WorkflowT',
        **kwargs,
    ):
        """
        Initializes the Workflow Context Component
        """
        self.workflow = workflow
        self.instance = workflow.instance
        self.hsettings: 'HatchetSettings' = workflow.hsettings
        self.workflow_name = workflow.workflow_name
        self.hatchet = workflow.hatchet
        self.name = 'workflow'
        
        # self.name = workflow.name

        self.configure_init(**kwargs)
        self.configure_pre_init(**kwargs)
        self.configure_post_init(**kwargs)

    @property
    def trace_started(self) -> bool:
        """
        Returns whether or not the trace started
        """
        if 'trace_started' not in self._extra:
            self._extra['trace_started'] = False
        return self._extra['trace_started']

    @trace_started.setter
    def trace_started(self, value: bool):
        """
        Sets the trace started
        """
        self._extra['trace_started'] = value

    @property
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        return self.hsettings.logger
    
    @property
    def autologger(self) -> 'Logger':
        """
        Gets the autologger
        """
        return self.hsettings.autologger

    def configure_init(self, *args, **kwargs):
        """
        Configure the Workerflow Component
        """
        pass

    def configure_pre_init(self, *args, **kwargs):
        """
        Configure the Workerflow Component
        """
        pass

    def configure_post_init(self, *args, **kwargs):
        """
        Post Configure the Workerflow Component
        """
        pass

    def get_process_state(self, prefix: Optional[str] = None, as_str: bool = True) -> Union[str, Dict[str, Any]]:
        """
        Captures the state of the process
        """
        if prefix: prefix = f'[{prefix}] '
        return capture_state(as_str = as_str, prefix = prefix, in_k8s = self.hsettings.in_k8s)

    def get_trace_state(self, prefix: str = 'Start', limit: int = 3) -> Optional[str]:
        """
        Gets the trace state
        """
        if prefix == 'Start': 
            if not self.trace_started:
                with contextlib.suppress(Exception):
                    tracemalloc.start()
                self.trace_started = True
            return
        s = display_top(tracemalloc.take_snapshot(), limit = limit)
        s = f'[{prefix}] {s}'
        return s


    @property
    def hatchet(self) -> 'HatchetSession':
        """
        Gets the Hatchet Client
        """
        self._set_client('hatchet')
        return self._clients['hatchet']
    
    @hatchet.setter
    def hatchet(self, value: 'HatchetSession'):
        """
        Sets the Hatchet Client
        """
        self._clients['hatchet'] = value
        if not self.instance: self.instance = value.instance


    async def on_workflow_error(self, context: 'Context', e: Exception, **kwargs):
        """
        Handles the workflow error
        """
        pass


    @contextlib.asynccontextmanager
    async def catch_workflow_error(
        self,
        context: 'Context',
        **kwargs,
    ):
        """
        Catches the workflow error and stores it in the cache
        """
        try:
            yield
        except Exception as e:
            await self.on_workflow_error(context, e)
            raise e
    

    def _set_client(self, name: str, key: Optional[str] = None) -> None:
        """
        Sets the client
        """
        key = key or name
        if key not in self._clients: self._clients[key] = self._get_client(name)

    def _get_client(self, name: str) -> Any:
        """
        Gets a client
        """
        raise NotImplementedError

    def _import(self, name: str) -> Any:
        """
        Imports a module
        """
        from lzl.load import lazy_import
        return lazy_import(name)
    
    def _get_kdb_session(self, name: str, **kwargs) -> 'KVDBSession':
        """
        Gets a KVDB Session
        """
        from kvdb import KVDBClient
        return KVDBClient.get_session(name = name, **kwargs)
    

    def _get_pdict(self, base_key: str, **kwargs) -> 'PersistentDict':
        """
        Gets a Persistent Dict
        """
        return self.pkdb.create_persistence(base_key = base_key, **kwargs)

    @property
    def kdb(self) -> 'KVDBSession':
        """
        Gets the KVDB Session
        """
        if 'kdb' not in self._clients:
            self._clients['kdb'] = self._get_kdb_session('workflow', serializer = 'json')
        return self._clients['kdb']

    @property
    def rkdb(self) -> 'KVDBSession':
        """
        Gets the Raw KVDB Session with Decoder Enabled
        """
        if 'rkvdb' not in self._clients:
            self._clients['rkvdb'] = self._get_kdb_session('workflow-raw', serializer = None, decode_responses = True)
        return self._clients['rkvdb']
    
    @property
    def pkdb(self) -> 'KVDBSession':
        """
        Gets the KVDB Session with no serialization for Persistence
        """
        if 'pkdb' not in self._clients:
            self._clients['pkdb'] = self._get_kdb_session('persistence', serializer = None)
        return self._clients['pkdb']

    @property
    def tmp_data(self) -> 'TemporaryData':
        """
        Gets the Temporary Data
        """
        return self.hsettings.temp_data


from __future__ import annotations

"""Utilities for building lazy-loading registries."""

import inspect
import os
from pathlib import Path
import typing as t

from lzl.proxied import ProxyObject
from lzo.utils.state import TempData

RT = t.TypeVar('RT')

__all__ = ['combine_parts', 'MRegistry']


def combine_parts(*parts: t.Optional[str], sep: str = '.') -> str:
    """Join non-empty fragments into a dotted identifier.

    Args:
        *parts: Sequence of string fragments that may include ``None`` values.
        sep: Delimiter inserted between fragments. Defaults to ``'.'``.

    Returns:
        A single string containing the provided fragments joined by ``sep``.
    """

    return sep.join(p for p in parts if p)


class MRegistry(t.Generic[RT]):
    """Mutable registry that lazily initialises and caches objects.

    The registry stores three parallel maps:

    - ``mregistry`` for classes/functions registered ahead of time.
    - ``uninit_registry`` for dotted import paths that are resolved on demand.
    - ``init_registry`` for concrete instances that have been constructed.

    Hooks can be attached to modify kwargs before instantiation (``prehooks``)
    or the resulting object after instantiation (``posthooks``).
    """

    mregistry: t.ClassVar[t.Dict[str, t.Type[RT]]] = {}
    uninit_registry: t.ClassVar[t.Dict[str, str]] = {}
    init_registry: t.ClassVar[t.Dict[str, RT]] = {}

    prehooks: t.ClassVar[t.Dict[str, t.Callable[..., t.Any]]] = {}
    posthooks: t.ClassVar[t.Dict[str, t.Callable[..., t.Any]]] = {}

    def __init__(
        self,
        name: str,
        verbose: t.Optional[bool] = False,
        **kwargs: t.Any,
    ) -> None:
        """Initialise the registry with a display name and verbosity flag.

        Args:
            name: Friendly name used when logging registration activity.
            verbose: Enables additional logging when objects are created.
            **kwargs: Forwarded for compatibility; unused presently.
        """

        from lzl.io.ser import get_object_classname
        from lzl.load import lazy_import
        from lzl.logging import logger

        self.name = name
        self.logger = logger
        self.get_classname = get_object_classname
        self.lazy_import = lazy_import
        self.verbose = verbose
        self.idx: t.Dict[str, RT] = {}
        self._extra: t.Dict[str, t.Any] = {}

    def _register(self, key: str, value: RT) -> None:
        """Store a class/constructor in the registry, replacing existing values."""

        self.mregistry[key] = value
        if key in self.uninit_registry:
            self.uninit_registry.pop(key)
        if os.getenv('MUTE_LZ_REGISTRY', 'false').lower() in {'true', '1'}:
            return
        if not isinstance(value, str) and getattr(value, '_rverbose', self.verbose):
            if not TempData.has_logged(f'lzo.registry.register:{key}'):
                self.logger.info(
                    f'Registered: {key}',
                    colored=True,
                    prefix=self.name,
                )

    def __setitem__(self, key: str, value: RT) -> None:
        """Alias for :meth:`_register` so the registry mimics a mapping."""

        self._register(key, value)

    def register_prehook(self, key: str, func: t.Union[t.Callable[..., t.Any], str]) -> None:
        """Attach a callable executed before an object is instantiated."""

        self.prehooks[key] = func

    def register_posthook(self, key: str, func: t.Union[t.Callable[..., t.Any], str]) -> None:
        """Attach a callable executed after an object is instantiated."""

        self.posthooks[key] = func

    def register_hook(
        self,
        key: str,
        func: t.Union[t.Callable[..., t.Any], str],
        kind: t.Literal['pre', 'post'] = 'pre',
    ) -> None:
        """Convenience wrapper for registering pre- or post-hooks."""

        if kind == 'pre':
            self.register_prehook(key, func)
        elif kind == 'post':
            self.register_posthook(key, func)

    def run_obj_init(
        self,
        key: str,
        obj: t.Union[t.Type[RT], RT],
        **kwargs: t.Any,
    ) -> RT:
        """Instantiate an object and invoke any configured hooks."""

        if key in self.prehooks:
            if isinstance(self.prehooks[key], str):
                self.prehooks[key] = self.lazy_import(self.prehooks[key])
            kwargs = self.prehooks[key](**kwargs)

        if isinstance(obj, ProxyObject):
            if self.verbose:
                self.logger.info(
                    f'Skipping Initialization for Proxy Object: {key}',
                    colored=True,
                    prefix=self.name,
                )
        else:
            obj = obj(**kwargs)

        if key in self.posthooks:
            if isinstance(self.posthooks[key], str):
                self.posthooks[key] = self.lazy_import(self.posthooks[key])
            obj = self.posthooks[key](obj)
        return obj

    def _register_initialized(self, key: str, value: RT) -> None:
        """Cache an already-instantiated object for repeated retrieval."""

        self.init_registry[key] = value

    def _get(
        self,
        key: str,
        _raise_error: bool = True,
        **kwargs: t.Any,
    ) -> RT:
        """Resolve the concrete object backing ``key`` without memoisation."""

        if key in self.init_registry:
            return self.init_registry[key]

        if key in self.uninit_registry:
            _path = self.uninit_registry[key]
            _obj = self.lazy_import(_path)
            self.init_registry[key] = self.run_obj_init(key, _obj, **kwargs)
            self.uninit_registry.pop(key, None)
            return self.init_registry[key]

        if key in self.mregistry:
            _obj = self.mregistry[key]
            self.init_registry[key] = self.run_obj_init(key, _obj, **kwargs)
            return self.init_registry[key]

        if not _raise_error:
            return None
        raise KeyError(f'Key {key} not found in {self.name}')

    def get(self, key: str, **kwargs: t.Any) -> RT:
        """Public accessor that memoises lookups for repeat calls."""

        if key in self.idx:
            return self.idx[key]
        if (item := self._get(key, _raise_error=False, **kwargs)) is not None:
            self.idx[key] = item
            return item
        if possible_key := self.search_for_parent(key, raise_error=False):
            if possible_key in self.idx:
                self.idx[key] = self.idx[possible_key]
                return self.idx[key]
            if (item := self._get(possible_key, _raise_error=False, **kwargs)) is not None:
                self.idx[key] = item
                return item
        raise KeyError(
            f'Key {key} not found in {self.name}: '
            f'init: `{list(self.init_registry.keys())}`, '
            f'idx: `{list(self.idx.keys())}`'
        )

    def get_module_path(self, obj: t.Type[RT]) -> Path:
        """Return the filesystem path where ``obj`` is defined."""

        return Path(inspect.getfile(obj)).parent

    def search_for_parent(self, key: str, raise_error: bool = True) -> t.Optional[str]:
        """Find a registry key that ends with ``key`` (supports partial lookups)."""

        if self.init_registry:
            for k in self.init_registry:
                if k.endswith(key):
                    return k

        if self.uninit_registry:
            for k in self.uninit_registry:
                if k.endswith(key):
                    return k

        if self.mregistry:
            for k in self.mregistry:
                if k.endswith(key):
                    return k

        if not raise_error:
            return None
        raise KeyError(f'Key {key} not found in {self.name}')

    def __getitem__(self, key: str) -> RT:
        """Allow bracket-notation access (``registry[key]``)."""

        return self.get(key)



from __future__ import annotations

"""
OpenAI Session Manager
"""

import abc
import copy
import pathlib
import random
from typing import Optional, List, Callable, Dict, Union, Any, overload, Tuple, TYPE_CHECKING, Literal
from lzl.types import eproperty
from lzl.proxied import proxied, ProxyObject
from lzl.api.openai.schemas import *
from lzl.api.openai.types.base import ApiType, Usage
# from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.configs import get_settings, OpenAISettings
from lzl.api.openai.configs.external import ExternalProviderSettings
from lzl.api.openai.utils import logger, weighted_choice
# from lzl.api.openai.types.functions import FunctionManager, OpenAIFunctions

from .loadbalancer import ClientLoadBalancer

if TYPE_CHECKING:
    from .base import OpenAIClient
    from .external import ExternalOpenAIClient
    from .functions import FunctionManager
    from lzl.api.openai.types.handlers import ModelCostHandler
    from lzl.pool import ThreadPool
    from lzl.api.openai.schemas.chat import Function



# Model Mapping for Azure
DefaultModelMapping = {
    'gpt-3.5-turbo': 'gpt-35-turbo',
    'gpt-3.5-turbo-16k': 'gpt-35-turbo-16k',
    'gpt-3.5-turbo-instruct': 'gpt-35-turbo-instruct',
    'gpt-3.5-turbo-0301': 'gpt-35-turbo-0301',
    'gpt-3.5-turbo-0613': 'gpt-35-turbo-0613',
    'gpt-3.5-turbo-1106': 'gpt-35-turbo-1106',
}
DefaultAvailableModels = [
    'gpt-4',
    'gpt-4-32k',
    'gpt-4-turbo',
    'gpt-4-1106-preview',
    'gpt-4-0125-preview',
    'text-embedding-ada-2',
    'text-embedding-3-small',
    'text-embedding-3-large',
] + list(DefaultModelMapping.values())

# @proxied 


class OpenAIManager(abc.ABC):
    name: Optional[str] = "openai"
    on_error: Optional[Callable] = None
    prioritize: Optional[str] = None
    auto_healthcheck: Optional[bool] = None
    auto_loadbalance_clients: Optional[bool] = None
    azure_model_mapping: Optional[Dict[str, str]] = DefaultModelMapping

    _api: Optional['OpenAIClient'] = None
    _apis: Optional['ClientLoadBalancer'] = None
    _clients: Optional[Dict[str, 'OpenAIClient']] = {}
    _settings: Optional[OpenAISettings] = None

    _pooler: Optional['ThreadPool'] = None

    """
    The Global Session Manager for OpenAI API.
    """

    def __init__(self, **kwargs):
        """
        Initializes the OpenAI API Client
        """
        self._extra: Dict[str, Any] = {}
        self.client_weights: Optional[Dict[str, float]] = {}
        self.client_ping_timeouts: Optional[Dict[str, float]] = {}
        self.client_model_exclusions: Optional[Dict[str, Dict[str, Union[bool, List[str]]]]] = {}
        self.client_base_urls: Optional[Dict[str, str]] = {}

        self.no_proxy_client_names: Optional[List[str]] = []
        self.client_callbacks: Optional[List[Callable]] = []
        # from .functions import OpenAIFunctions
        # self.functions: 'FunctionManager' = OpenAIFunctions
        # self.ctx = ModelContextHandler
        if self.auto_loadbalance_clients is None: self.auto_loadbalance_clients = self.settings.auto_loadbalance_clients
        if self.auto_healthcheck is None: self.auto_healthcheck = self.settings.auto_healthcheck

        self.external_clients: Dict[str, 'ExternalOpenAIClient'] = {}
        self.external_client_weights: Optional[Dict[str, float]] = {}
        self.external_model_to_client: Dict[str, str] = {}
        self.external_client_default: Optional[str] = None
        # cn = f'{self.__class__.__module__}.{self.__class__.__qualname__}'
        # logger.info(f'initizing OpenAIClient - {cn}')
        # if 'lzl.api' in cn:
        #     raise RuntimeError('Cannot initialize OpenAIClient from `lzl.api`. Please use `lzl.api.openai.client.OpenAIClient` instead.')
        # self._external_clients: 

    @property
    def has_client_weights(self) -> bool:
        """
        Returns if the client has weights
        """
        return bool(self.client_weights)
    
    @property
    def has_external_client_weights(self) -> bool:
        """
        Returns if the client has weights
        """
        return bool(self.external_client_weights)

    def add_callback(self, callback: Callable):
        """
        Adds a callback to the client
        """
        self.client_callbacks.append(callback)

    @property
    def settings(self) -> OpenAISettings:
        """
        Returns the global settings for the OpenAI API.
        """
        if self._settings is None:
            self._settings = get_settings()
        return self._settings
    
    @eproperty
    def functions(self) -> 'FunctionManager':
        """
        Returns the global functions manager
        """
        from .functions import OpenAIFunctions
        return OpenAIFunctions

    @eproperty
    def ctx(self) -> 'ModelCostHandler':
        """
        Returns the global model context handler
        """
        from lzl.api.openai.types.handlers import ModelContextHandler
        return ModelContextHandler

    
    # Changing the behavior to become proxied through settings

    @property
    def api_key(self) -> Optional[str]:
        """
        Returns the global API Key.
        """
        return self.settings.api_key
    
    @property
    def url(self) -> Optional[str]:
        """
        Returns the global URL.
        """
        return self.settings.url
    
    @property
    def scheme(self) -> Optional[str]:
        """
        Returns the global Scheme.
        """
        return self.settings.scheme
    
    @property
    def host(self) -> Optional[str]:
        """
        Returns the global Host.
        """
        return self.settings.host
    
    @property
    def port(self) -> Optional[int]:
        """
        Returns the global Port.
        """
        return self.settings.port
    
    @property
    def api_base(self) -> Optional[str]:
        """
        Returns the global API Base.
        """
        return self.settings.api_base

    @property
    def api_path(self) -> Optional[str]:
        """
        Returns the global API Path.
        """
        return self.settings.api_path
    
    @property
    def api_type(self) -> Optional[ApiType]:
        """
        Returns the global API Type.
        """
        return self.settings.api_type
    
    @property
    def api_version(self) -> Optional[str]:
        """
        Returns the global API Version.
        """
        return self.settings.api_version
    
    @property
    def api_key_path(self) -> Optional[pathlib.Path]:
        """
        Returns the global API Key Path.
        """
        return self.settings.api_key_path
    
    @property
    def organization(self) -> Optional[str]:
        """
        Returns the global Organization.
        """
        return self.settings.organization
    
    @property
    def proxies(self) -> Optional[Union[str, Dict]]:
        """
        Returns the global Proxies.
        """
        return self.settings.proxies
    
    @property
    def timeout(self) -> Optional[int]:
        """
        Returns the global Timeout.
        """
        return self.settings.timeout
    
    @property
    def max_retries(self) -> Optional[int]:
        """
        Returns the global Max Retries.
        """
        return self.settings.max_retries
    
    @property
    def app_info(self) -> Optional[Dict[str, str]]:
        """
        Returns the global App Info.
        """
        return self.settings.app_info
    
    @property
    def debug_enabled(self) -> Optional[bool]:
        """
        Returns the global Debug Enabled.
        """
        return self.settings.debug_enabled
    
    @property
    def ignore_errors(self) -> Optional[bool]:
        """
        Returns the global Ignore Errors.
        """
        return self.settings.ignore_errors
    
    @property
    def timeout(self) -> Optional[int]:
        """
        Returns the global Timeout.
        """
        return self.settings.timeout
    
    @property
    def pooler(self) -> Optional['ThreadPool']:
        """
        Returns the global ThreadPool.
        """
        if self._pooler is None:
            from lzl.pool import ThreadPool
            self._pooler = ThreadPool
        return self._pooler

    
    def configure_client(
        self,
        client_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Configure a specific client.
        """
        if self.auto_loadbalance_clients:
            return self.apis.configure_client(client_name = client_name, **kwargs)
        client_name = client_name or 'default'
        if client_name not in self._clients:
            raise ValueError(f'Client `{client_name}` does not exist.')
        self._clients[client_name].reset(**kwargs)

    def get_api_client(
        self,
        client_name: Optional[str] = None,
        **kwargs,
    ) -> 'OpenAIClient':
        """
        Initializes a new OpenAI client or Returns an existing one.
        """
        if self.auto_loadbalance_clients:
            return self.apis.get_api_client(client_name = client_name, **kwargs)
        client_name = client_name or 'default'
        if client_name not in self._clients:
            self._clients[client_name] = self.init_api_client(client_name = client_name, **kwargs)
        return self._clients[client_name]
    

    def get_api_client_from_list(
        self,
        client_names: Optional[List[str]] = None,
        **kwargs,
    ) -> 'OpenAIClient':
        """
        Initializes a new OpenAI client or Returns an existing one.
        """
        if self.auto_loadbalance_clients:
            if not client_names: return self.apis.get_api_client(**kwargs)
            return self.apis.get_api_client_from_list(client_names = client_names, **kwargs)
        if not client_names: return self.get_api_client(**kwargs)
        
        if not self.auto_healthcheck:
            name = self.select_client_name_from_weights(client_names) if self.has_client_weights else random.choice(client_names)
            return self.get_api_client(client_name = name, **kwargs)
        
        available = []
        for client_name in client_names:
            if client_name not in self._clients:
                self._clients[client_name] = self.init_api_client(client_name = client_name, **kwargs)
            if not self._clients[client_name].ping(**self.get_client_ping_params(client_name)):
                continue
            if not self.has_client_weights:
                return self._clients[client_name]
            available.append(client_name)
        
        if available:
            name = self.select_client_name_from_weights(available)
            return self._clients[name]
        raise ValueError(f'No healthy client found from: {client_names}')
    
    async def aget_api_client_from_list(
        self,
        client_names: Optional[List[str]] = None,
        **kwargs,
    ) -> 'OpenAIClient':
        """
        Initializes a new OpenAI client or Returns an existing one.
        """
        if self.auto_loadbalance_clients:
            if not client_names: return self.apis.get_api_client(**kwargs)
            return await self.apis.aget_api_client_from_list(client_names = client_names, **kwargs)
        if not client_names: return self.get_api_client(**kwargs)
        if not self.auto_healthcheck:
            name = self.select_client_name_from_weights(client_names) if self.has_client_weights else random.choice(client_names)
            return self.get_api_client(client_name = name, **kwargs)
        
        available = []
        for client_name in client_names:
            if client_name not in self._clients:
                self._clients[client_name] = self.init_api_client(client_name = client_name, **kwargs)
            if not await self._clients[client_name].aping(**self.get_client_ping_params(client_name)):
                continue
            if not self.has_client_weights:
                return self._clients[client_name]
            available.append(client_name)
        
        if available:
            name = self.select_client_name_from_weights(available)
            return self._clients[name]
        raise ValueError(f'No healthy client found from: {client_names}')
    
        
    def init_api_client(
        self, 
        client_name: Optional[str] = None,
        set_as_default: Optional[bool] = False,
        is_azure: Optional[bool] = None,
        **kwargs
    ) -> 'OpenAIClient':
        """
        Creates a new OpenAI client.
        """
        if self.auto_loadbalance_clients:
            return self.apis.init_api_client(client_name = client_name, set_as_default = set_as_default, is_azure = is_azure, **kwargs)
        client_name = client_name or 'default'
        if client_name in self._clients:
            return self._clients[client_name]
        
        from .base import OpenAIClient
        if is_azure is None and \
            (
                # (client_name == 'default' or 'az' in client_name) and 
                'az' in client_name and self.settings.has_valid_azure
            ):
            is_azure = True
        if 'client_callbacks' not in kwargs and self.client_callbacks:
            kwargs['client_callbacks'] = self.client_callbacks
        client = OpenAIClient(
            name = client_name,
            settings = self.settings,
            is_azure = is_azure,
            azure_model_mapping = self.azure_model_mapping,
            **kwargs
        )
        self._clients[client_name] = client
        if set_as_default or not self._api:
            self._api = client
        return client
    
    def configure_external_client(
        self,
        client: 'ExternalOpenAIClient',
        set_as_default: Optional[bool] = False,
        **kwargs,
    ):
        """
        Configures the External Client
        """
        self.external_clients[client.name] = client
        if set_as_default or not self.external_client_default:
            self.external_client_default = client.name
        if client.provider.config.weight is not None:
            self.external_client_weights[client.name] = client.provider.config.weight
        for model_name in client.provider.model_list:
            provider_model_name = f'{client.name}/{model_name}'
            self.external_model_to_client[provider_model_name] = client.name
            if model_name not in self.external_model_to_client or set_as_default:
                self.external_model_to_client[model_name] = client.name
        
        # logger.info(f'Configured External Models: {list(self.external_model_to_client.keys())}')

    
    def init_external_api_client(
        self,
        provider: ExternalProviderSettings,
        client_name: Optional[str] = None,
        set_as_default: Optional[bool] = False,
        disable_proxy: Optional[bool] = None,
        **kwargs
    ) -> 'ExternalOpenAIClient':
        """
        Initializes an external OpenAI client
        """
        client_name = client_name or provider.name
        if client_name in self.external_clients:
            return self.external_clients[client_name]
        from .external import ExternalOpenAIClient
        # from async_openai.external_client import ExternalOpenAIClient
        if 'client_callbacks' not in kwargs and self.client_callbacks:
            kwargs['client_callbacks'] = self.client_callbacks
        extra = ''
        if not disable_proxy and provider.config.has_proxy:
            # Configure the proxy version of the client
            client = ExternalOpenAIClient(
                name = client_name,
                provider = provider,
                settings = self.settings,
                is_proxied = True,
                **kwargs
            )
            self.configure_external_client(client, set_as_default = set_as_default)
            # Initialize a non-proxy version of the client
            non_proxy_client_name = f'{client_name}_noproxy'
            non_proxy_client = ExternalOpenAIClient(
                name = non_proxy_client_name,
                provider = provider,
                settings = self.settings,
                is_proxied = False,
                **kwargs
            )
            self.configure_external_client(non_proxy_client, set_as_default = False)
            extra += ' Proxied,'
        else:
            client = ExternalOpenAIClient(
                name = client_name,
                provider = provider,
                settings = self.settings,
                is_proxied = False,
                **kwargs
            )
            self.configure_external_client(client, set_as_default = set_as_default)
        if client.provider.config.has_api_keys:
            extra += f' Multiple API Keys: {len(client.provider.config.api_keys)},'
        extra = extra.strip().rstrip(',')
        if extra: extra = f', {extra}'
        logger.info(f'Registered `|g|{client.name}|e|` @ `{client.provider.config.api_url}` (External: |g|{client.provider.name}|e|{extra})', colored = True)
        return client


    def init_external_api_client_from_preset(
        self,
        preset_name: Optional[str] = None,
        preset_path: Optional[Union[str, pathlib.Path]] = None,
        client_name: Optional[str] = None,
        set_as_default: Optional[bool] = False,
        disable_proxy: Optional[bool] = None,
        overrides: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> 'ExternalOpenAIClient':
        """
        Initializes an external OpenAI client from a preset
        """
        overrides = overrides or {}
        provider = ExternalProviderSettings.from_preset(name = preset_name, path = preset_path, **overrides)
        return self.init_external_api_client(
            provider = provider, 
            client_name = client_name, 
            set_as_default = set_as_default,
            disable_proxy = disable_proxy, 
            **kwargs
        )


    def rotate_client(self, index: Optional[int] = None, verbose: Optional[bool] = False, **kwargs):
        """
        Rotates the clients
        """
        if not self.auto_loadbalance_clients:
            raise ValueError('Rotating Clients is not enabled.')
        self.apis.rotate_client(index = index, verbose = verbose, **kwargs)
    
    def set_client(self, client_name: Optional[str] = None, verbose: Optional[bool] = False):
        """
        Sets the client
        """
        if self.auto_loadbalance_clients:
            self.apis.set_client(client_name = client_name, verbose = verbose)
        else:
            self._api = self._clients[client_name]
            if verbose:
                logger.info(f'Set Client: {self.api.name} ({self.api.is_azure})')
    
    def get_current_client_info(self, verbose: Optional[bool] = False) -> Dict[str, Union[str, int]]:
        """
        Returns the current client info
        """
        if self.auto_loadbalance_clients:
            return self.apis.current_client_info(verbose = verbose)
        data = {
            'name': self.api.name,
            'is_azure': self.api.is_azure,
            'api_version': self.api.api_version,
        }
        if verbose:
            logger.info(f'Current Client: {self.api.name} (Azure: {self.api.is_azure} - {self.api.api_version})')
        return data

    
    @property
    def apis(self) -> ClientLoadBalancer:
        """
        Returns the global Rotating Clients.
        """
        if self._apis is None:
            self._apis = ClientLoadBalancer(
                prioritize=self.prioritize, 
                settings=self.settings, 
                azure_model_mapping=self.azure_model_mapping, 
                healthcheck=self.auto_healthcheck,
                manager = self,
            )
            if self.settings.client_configurations:
                self.register_client_endpoints()
            else:
                self.register_default_endpoints()
        return self._apis
    
    @property
    def api(self) -> 'OpenAIClient':
        """
        Returns the inherited OpenAI client.
        """
        if self.auto_loadbalance_clients: return self.apis.api
        if self._api is None:
            self.init_api_client()
        return self._api
    
    @property
    def external_api(self) -> 'ExternalOpenAIClient':
        """
        Returns the inherited External OpenAI client.
        """
        if not self.external_client_default or not self.external_clients:
            raise ValueError('No External Client has been initialized.')
        return self.external_clients[self.external_client_default]
    
    def configure_internal_apis(self):
        """
        Helper method to ensure that the APIs are initialized
        """
        if self._api is not None: return
        # Invoke it to ensure that it is initialized
        if self.auto_loadbalance_clients:
            self.apis
        else:
            self.init_api_client()

    def _ensure_api(self):
        """
        Ensures that the API is initialized
        """
        if self._api is None: self.configure_internal_apis()

    def select_client_name_from_weights(self, names: List[str]) -> str:
        """
        Returns the client weights
        """
        return weighted_choice([(name, self.client_weights.get(name, 1.0)) for name in names])

    def select_external_client_name_from_weights(self, names: List[str]) -> str:
        """
        Returns the client weights
        """
        return weighted_choice([(name, self.external_client_weights.get(name, 1.0)) for name in names])

    # def get_client_ping_timeout(self, name: str) -> Optional[float]:
    #     """
    #     Returns the client timeout
    #     """
    #     return self.client_ping_timeouts.get(name, 1.0)
    
    def get_client_ping_params(self, name: str) -> Dict[str, Union[float, str]]:
        """
        Returns the client ping parameters
        """
        return {
            'timeout': self.client_ping_timeouts.get(name, 1.0),
            'base_url': self.client_base_urls.get(name),
        }

    """
    API Routes
    """

    @property
    def completions(self) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return self.api.completions
    
    @property
    def Completions(self) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return self.api.completions
    

    @property
    def chat(self) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat`.
        
        Doc: `https://beta.openai.com/docs/api-reference/chat`
        """
        return self.api.chat
    
    @property
    def Chat(self) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat`.
        
        Doc: `https://beta.openai.com/docs/api-reference/chat`
        """
        return self.api.chat
    
    # @property
    # def edits(self) -> EditRoute:
    #     """
    #     Returns the `EditRoute` class for interacting with `Edits`.
        
    #     Doc: `https://beta.openai.com/docs/api-reference/edits`
    #     """
    #     return self.api.edits
    
    @property
    def embeddings(self) -> EmbeddingRoute:
        """
        Returns the `EmbeddingRoute` class for interacting with `Embeddings`.
        
        Doc: `https://beta.openai.com/docs/api-reference/embeddings`
        """
        return self.api.embeddings

    @property
    def rerankings(self) -> RerankingRoute:
        """
        Returns the `RerankingRoute` class for interacting with `Rerankings`.
        
        Doc: `https://beta.openai.com/docs/api-reference/rerankings`
        """
        return self.api.rerankings
    
    # @property
    # def images(self) -> ImageRoute:
    #     """
    #     Returns the `ImageRoute` class for interacting with `Images`.
        
    #     Doc: `https://beta.openai.com/docs/api-reference/images`
    #     """
    #     return self.api.images
    
    @property
    def models(self) -> ModelRoute:
        """
        Returns the `ModelRoute` class for interacting with `models`.
        
        Doc: `https://beta.openai.com/docs/api-reference/models`
        """
        return self.api.models


    """
    V2 Endpoint Registration with Proxy Support
    """

    def register_default_endpoints(self):
        """
        Register the default clients
        """
        if self.settings.proxy.enabled:
            api_base = self.settings.proxy.endpoint
            az_custom_headers = {
                "Helicone-OpenAI-Api-Base": self.settings.azure.api_base
            }
            self.configure(
                api_base = api_base,
                azure_api_base = api_base,
                azure_custom_headers = az_custom_headers,
                enable_rotating_clients = True,
                prioritize = "azure",
            )

        self.init_api_client('openai', is_azure = False)
        if self.settings.has_valid_azure:
            self.init_api_client('azure', is_azure = True)


    def register_client_endpoints(self):  # sourcery skip: low-code-quality
        """
        Register the Client Endpoints
        """
        client_configs = copy.deepcopy(self.settings.client_configurations)
        seen_models = set(DefaultAvailableModels)
        has_weights = any(c.get('weight') for c in client_configs.values())
        for name, config in client_configs.items():
            is_enabled = config.pop('enabled', False)
            if not is_enabled: continue
            is_azure = 'azure' in name or 'az' in name or config.get('is_azure', False)
            is_default = config.pop('default', False)
            proxy_disabled = config.pop('proxy_disabled', False)
            source_endpoint = config.get('api_base')
            client_weight = config.pop('weight', 1.0) if has_weights else None
            client_ping_timeout = config.pop('ping_timeout', None)

            if self.debug_enabled is not None: config['debug_enabled'] = self.debug_enabled
            if excluded_models := config.pop('excluded_models', None):
                self.client_model_exclusions[name] = {
                    'models': excluded_models, 'is_azure': is_azure,
                }
                seen_models.update(excluded_models)
            else:
                self.client_model_exclusions[name] = {
                    'models': None, 'is_azure': is_azure,
                }
            
            if included_models := config.pop('included_models', None):
                self.client_model_exclusions[name]['included_models'] = included_models
            
            if client_weight: self.client_weights[name] = client_weight
            if client_ping_timeout is not None: self.client_ping_timeouts[name] = client_ping_timeout
            _has_proxy = False
            _proxy_provider = None
            if (self.settings.proxy.enabled and not proxy_disabled) and config.get('api_base'):
                # Initialize a non-proxy version of the client
                config['api_base'] = source_endpoint
                non_proxy_name = f'{name}_noproxy'
                self.client_base_urls[name] = source_endpoint
                if client_weight: self.client_weights[non_proxy_name] = client_weight
                if client_ping_timeout is not None: self.client_ping_timeouts[non_proxy_name] = client_ping_timeout
                self.client_model_exclusions[non_proxy_name] = self.client_model_exclusions[name].copy()
                self.no_proxy_client_names.append(non_proxy_name)
                self.init_api_client(non_proxy_name, is_azure = is_azure, set_as_default = False, **config)
                config['headers'] = self.settings.proxy.create_proxy_headers(
                    name = name,
                    config = config,
                )
                config['api_base'] = self.settings.proxy.endpoint
                _has_proxy = True
                _proxy_provider = self.settings.proxy.current
            c = self.init_api_client(name, is_azure = is_azure, set_as_default = is_default, **config)
            msg = f'Registered: `|g|{c.name}|e|` @ `{source_endpoint or c.base_url}`'
            extra_msgs = []
            if has_weights: 
                if isinstance(client_weight, float):
                    _wp, _wsfx = '|g|', '|e|'
                    if client_weight <= 0.0: 
                        _wp, _wsfx = '', ''
                    elif client_weight <= 0.25: _wp = '|r|'
                    elif client_weight <= 0.45: _wp = '|y|'
                    extra_msgs.append(f'Weight: {_wp}{client_weight}{_wsfx}')
                else:
                    extra_msgs.append(f'Weight: {client_weight}')
            if c.is_azure: extra_msgs.append('Azure')
            if _has_proxy: extra_msgs.append(f'Proxied (via |g|{_proxy_provider}|e|)')
            if extra_msgs: msg += f' ({", ".join(extra_msgs)})'
            logger.info(msg, colored = True)
        
        # Set the models for inclusion
        for name in self.client_model_exclusions:
            if not self.client_model_exclusions[name].get('included_models'): continue
            included_models = self.client_model_exclusions[name].pop('included_models')
            self.client_model_exclusions[name]['models'] = [m for m in seen_models if m not in included_models]
            # if self.settings.debug_enabled:
            # logger.info(f'|g|{name}|e| Included: {included_models}, Excluded: {self.client_model_exclusions[name]["models"]}', colored = True)


    def select_client_names(
        self, 
        client_name: Optional[str] = None, 
        azure_required: Optional[bool] = None, 
        openai_required: Optional[bool] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
    ) -> Optional[List[str]]:
        """
        Select Client based on the client name, azure_required, and model
        """
        if client_name is not None: return client_name
        oai_name = 'openai_noproxy' if noproxy_required else 'openai'
        if openai_required: return [oai_name]
        if model is not None:
            available_clients = []
            for name, values in self.client_model_exclusions.items():
                if (noproxy_required and 'noproxy' not in name) or (not noproxy_required and 'noproxy' in name): continue
                if excluded_clients and name in excluded_clients: continue

                # Prioritize Azure Clients
                if (
                    azure_required and not values['is_azure']
                ) or (
                    not azure_required and not values['is_azure']
                ):
                    continue
                if not values['models'] or model not in values['models']:
                    available_clients.append(name)
                    # return name

            if not available_clients: available_clients.append(oai_name)
            return available_clients
        if azure_required:
            return [
                k for k, v in self.client_model_exclusions.items() if v['is_azure'] and \
                    ('noproxy' in k if noproxy_required else 'noproxy' not in k)
            ]
            # return [k for k, v in self.client_model_exclusions.items() if v['is_azure']]
        # Only 1 client / likely openai
        if len(self.client_model_exclusions) == 2:
            return (
                [i for i in self.client_model_exclusions if 'noproxy' in i]
                if noproxy_required
                else [
                    i for i in self.client_model_exclusions if 'noproxy' not in i
                ]
            )
        return None
    
    @property
    def external_client_names(self) -> List[str]:
        """
        Returns the list of external client names
        """
        return list(self.external_clients.keys())
    
    def select_external_client_names(
        self,
        client_name: Optional[str] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
    ) -> Optional[List[str]]:
        """
        Select External Client based on the client name, and model
        """
        if client_name and model:
            client_names = [
                client for model_name, client in self.external_model_to_client.items() if model in model_name \
                and client_name in model_name
            ]
        elif client_name: client_names = [name for name in self.external_client_names if client_name in name]
        elif model: client_names = [client for model_name, client in self.external_model_to_client.items() if model in model_name]
        else: client_names = self.external_client_names
        if noproxy_required: client_names = [c for c in client_names if 'noproxy' in c]
        else: client_names = [c for c in client_names if 'noproxy' not in c]
        # logger.info(f'External Clients: {client_names}')
        if excluded_clients: client_names = [c for c in client_names if c not in excluded_clients]
        return list(set(client_names))

    
    def get_client(
        self, 
        client_name: Optional[str] = None, 
        azure_required: Optional[bool] = None, 
        openai_required: Optional[bool] = None,
        # provider_required: Optional[str] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs,
    ) -> Union["OpenAIClient", "ExternalOpenAIClient"]:
        """
        Gets the OpenAI client

        Args:
            client_name (str, optional): The name of the client to use. If not provided, it will be selected based on the other parameters.
            azure_required (bool, optional): Whether the client must be an Azure client.
            openai_required (bool, optional): Whether the client must be an OpenAI client.
            model (str, optional): The model to use. If provided, the client will be selected based on the model.
            noproxy_required (bool, optional): Whether the client must be a non-proxy client.
            excluded_clients (List[str], optional): A list of client names to exclude from selection.
        """
        self._ensure_api()
        # print(self.external_model_to_client, model, self.external_client_names)
        if (
            model and ('/' in model or model in self.external_model_to_client)
        ) or (
            client_name and client_name in self.external_clients
        ):
            # print('getting exeternal client')
            return self.get_external_client(client_name = client_name, model = model, noproxy_required = noproxy_required, excluded_clients = excluded_clients, **kwargs)
        client_names = self.select_client_names(
            client_name = client_name, 
            azure_required = azure_required, 
            openai_required = openai_required, 
            model = model, 
            noproxy_required = noproxy_required, 
            excluded_clients = excluded_clients
        )
        client = self.get_api_client_from_list(
            client_names = client_names, 
            azure_required = azure_required,
            **kwargs
        )
        if self.debug_enabled:
            logger.info(f'Available Clients: {client_names} - Selected: `|g|{client.name}|e|`', colored = True)
        if not client_name and self.auto_loadbalance_clients:
            self.apis.increase_rotate_index()
        return client
    
    def get_external_client(
        self,
        client_name: Optional[str] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs,
    ) -> "ExternalOpenAIClient":
        """
        Gets the External OpenAI client
        """
        client_names = self.select_external_client_names(
            client_name = client_name, 
            model = model, 
            noproxy_required = noproxy_required, 
            excluded_clients = excluded_clients
        )
        client_name = self.select_external_client_name_from_weights(client_names) if self.has_external_client_weights else random.choice(client_names)
        client = self.external_clients[client_name]
        if self.debug_enabled:
            logger.info(f'Available Clients: {client_names} - Selected: {client.name}')
        return client
    

    def get_chat_client(
        self, 
        client_name: Optional[str] = None, 
        azure_required: Optional[bool] = None, 
        openai_required: Optional[bool] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs,
    ) -> ChatRoute:
        """
        Gets the chat client
        """
        return self.get_client(client_name = client_name, azure_required = azure_required, openai_required = openai_required, model = model, noproxy_required = noproxy_required, excluded_clients = excluded_clients, **kwargs).chat

    def get_completion_client(
        self, 
        client_name: Optional[str] = None, 
        azure_required: Optional[bool] = None, 
        openai_required: Optional[bool] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs
    ) -> CompletionRoute:
        """
        Gets the chat client
        """
        return self.get_client(client_name = client_name, azure_required = azure_required, openai_required = openai_required, model = model, noproxy_required = noproxy_required, excluded_clients = excluded_clients, **kwargs).completions
    
    def get_embedding_client(
        self, 
        client_name: Optional[str] = None, 
        azure_required: Optional[bool] = None, 
        openai_required: Optional[bool] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs
    ) -> EmbeddingRoute:
        """
        Gets the embedding client
        """
        return self.get_client(client_name = client_name, azure_required = azure_required, openai_required = openai_required, model = model, noproxy_required = noproxy_required, excluded_clients = excluded_clients, **kwargs).embeddings


    def get_reranking_client(
        self, 
        client_name: Optional[str] = None, 
        # azure_required: Optional[bool] = None, 
        # openai_required: Optional[bool] = None,
        # provider_required: Optional[str] = None,
        model: Optional[str] = None,
        noproxy_required: Optional[bool] = None,
        excluded_clients: Optional[List[str]] = None,
        **kwargs
    ) -> RerankingRoute:
        """
        Gets the reranking client
        """
        return self.get_client(
            client_name = client_name, 
            # azure_required = azure_required, 
            # openai_required = openai_required, 
            # provider_required = provider_required,
            model = model, 
            noproxy_required = noproxy_required, 
            excluded_clients = excluded_clients, 
            **kwargs
        ).rerankings


    """
    V2 Utilities
    """
    def truncate_to_max_length(
        self,
        text: str,
        model: str,
        max_length: Optional[int] = None,
        buffer_length: Optional[int] = None,
    ) -> str:
        """
        Truncates the text to the max length
        """
        if max_length is None:
            model_ctx = self.ctx.get(model)
            max_length = model_ctx.context_length
        if buffer_length is not None: max_length -= buffer_length
        
        encoder = self.ctx.get_tokenizer(model)
        tokens = encoder.encode(text)
        if len(tokens) > max_length:
            tokens = tokens[-max_length:] # type: ignore
            decoded = encoder.decode(tokens)
            text = text[-len(decoded):]
        return text
    
    def truncate_batch_to_max_length(
        self,
        texts: List[str],
        model: str,
        max_length: Optional[int] = None,
        buffer_length: Optional[int] = None,
    ) -> List[str]:
        """
        Truncates the text to the max length
        """
        if max_length is None:
            model_ctx = self.ctx.get(model)
            max_length = model_ctx.context_length
        if buffer_length is not None: max_length -= buffer_length
        encoder = self.ctx.get_tokenizer(model)
        truncated_texts = []
        for text in texts:
            tokens = encoder.encode(text)
            if len(tokens) > max_length:
                tokens = tokens[-max_length:] # type: ignore
                decoded = encoder.decode(tokens)
                text = text[-len(decoded):]
            truncated_texts.append(text)
        return truncated_texts

    async def atruncate_to_max_length(
        self,
        text: str,
        model: str,
        max_length: Optional[int] = None,
        buffer_length: Optional[int] = None,
    ) -> str:
        """
        Truncates the text to the max length
        """
        return await self.pooler.arun(
            self.truncate_to_max_length,
            text = text,
            model = model,
            max_length = max_length,
            buffer_length = buffer_length,
        )
    
    async def atruncate_batch_to_max_length(
        self,
        texts: List[str],
        model: str,
        max_length: Optional[int] = None,
        buffer_length: Optional[int] = None,
    ) -> List[str]:
        """
        Truncates the text to the max length
        """
        return await self.pooler.arun(
            self.truncate_batch_to_max_length,
            texts = texts,
            model = model,
            max_length = max_length,
            buffer_length = buffer_length,
        )
            


    """
    Context Managers
    """

    async def async_close(self):
        """
        Closes the OpenAI API Client.
        """
        for client in self._clients.values():
            await client.async_close()

    
    def close(self):
        """
        Closes the OpenAI API Client.
        """
        for client in self._clients.values():
            client.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.async_close()

    def __getitem__(self, key: Union[int, str]) -> Union['OpenAIClient', 'ExternalOpenAIClient']:
        """
        Returns the OpenAI API Client.
        """
        if isinstance(key, str) and key in self.external_clients:
            return self.external_clients[key]
        if self.auto_loadbalance_clients:
            return self.apis[key]
        if isinstance(key, int):
            key = self.client_names[key]
        return self._clients[key]
    
    @property
    def client_names(self) -> List[str]:
        """
        Returns the list of client names.
        """
        return list(self._clients.keys())


    """
    Auto Rotating Functions
    """

    def chat_create(
        self, 
        input_object: Optional[ChatObject] = None,
        parse_stream: Optional[bool] = True,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        verbose: Optional[bool] = False,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.chat_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096 / 8182 / 32,768). If max_tokens is not provided, the model will use the maximum number of tokens
        Default: None

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): CURRENTLY NOT SUPPORTED
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :functions (optional): A list of dictionaries representing the functions to call
        
        :function_call (optional): The name of the function to call. Default: `auto` if functions are provided

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `ChatResponse`
        """
    
        try:
            return self.api.chat.create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)

        except Exception as e:
            if not self.auto_loadbalance_clients: raise e
            self.rotate_client(verbose=verbose)
            return self.chat_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)
    
    @overload
    async def async_chat_create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, str]]],
        model: Optional[str] = None,
        functions: Optional[List['Function']] = None,
        function_call: Optional[Union[str, Dict[str, str]]] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.chat.async_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    functions = [{'name': 'test', 'parameters': {'test': 'test'}}],
        >>>    function_call = 'auto'
        >>> )
        ```
        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`

        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.

        :temperature (optional): What sampling temperature to use. Higher values means

        :functions (optional): A list of dictionaries representing the functions to call

        :function_call (optional): The name of the function to call

        :auto_retry (optional): Whether to automatically retry the request if it fails

        :auto_retry_limit (optional): The maximum number of times to retry the request

        :header_cache_keys (optional): A list of keys to use for caching the headers

        Returns a ChatResponse
        """
        ...

    async def async_chat_create(
        self, 
        input_object: Optional[ChatObject] = None,
        parse_stream: Optional[bool] = True,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        verbose: Optional[bool] = False,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.async_chat_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional):
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.

        :functions (optional): A list of dictionaries representing the functions to call
        
        :function_call (optional): The name of the function to call. Default: `auto` if functions are provided

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Default: `None`

        Returns: `ChatResponse`
        """
        if 'model' in kwargs:
            client = self.get_chat_client(**kwargs)
            return await client.async_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)
        try:
            return await self.api.chat.async_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)

        except Exception as e:
            if not self.auto_loadbalance_clients: raise e
            self.rotate_client(verbose=verbose)
            return await self.async_chat_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)
    
    achat_create = async_chat_create

    def create_embeddings(
        self,
        inputs: Union[str, List[str]],
        model: Optional[str] = None,
        auto_retry: Optional[bool] = True,
        strip_newlines: Optional[bool] = False,
        headers: Optional[Dict[str, str]] = None,
        noproxy_required: Optional[bool] = False,
        **kwargs,
    ) -> List[List[float]]:
        """
        Creates the embeddings

        Args:
            inputs (Union[str, List[str]]): The input text or list of input texts.
            model (str, optional): The model to use. Defaults to None.
            auto_retry (bool, optional): Whether to automatically retry the request. Defaults to True.
            strip_newlines (bool, optional): Whether to strip newlines from the input. Defaults to False.
        """
        from lzo.utils.helpers.batching import split_into_batches
        model = model or 'text-embedding-ada-002'
        client = self.get_client(model = model, noproxy_required = noproxy_required, **kwargs)

        inputs = [inputs] if isinstance(inputs, str) else inputs
        if client.supports_tokenization:
            inputs = self.truncate_batch_to_max_length(
                inputs, 
                model = model, 
                **kwargs
            )
        
        if strip_newlines: inputs = [i.replace('\n', ' ').strip() for i in inputs]
        if not client.is_azure:
            response = client.embeddings.create(input = inputs, model = model, auto_retry = auto_retry, headers = headers, **kwargs)
            return response.embeddings

        embeddings = []
        # We need to split into batches of 5 for Azure
        # Azure has a limit of 5 inputs per request
        batches = split_into_batches(inputs, 5)
        for batch in batches:
            response = client.embeddings.create(input = batch, model = model, auto_retry = auto_retry, headers = headers, **kwargs)
            embeddings.extend(response.embeddings)
            # Shuffle the clients to load balance
            client = self.get_client(model = model, azure_required = True, noproxy_required = noproxy_required, **kwargs)
        return embeddings

    @overload
    async def async_create_embeddings(
        self,
        inputs: Union[str, List[str]],
        model: Optional[str] = ...,
        auto_retry: Optional[bool] = ...,
        strip_newlines: Optional[bool] = ...,
        headers: Optional[Dict[str, str]] = ...,
        noproxy_required: Optional[bool] = ...,
        include_metadata: Literal[False] = ...,
        **kwargs,
    ) -> List[List[float]]:
        """
        Creates the embeddings

        Args:
            inputs (Union[str, List[str]]): The input text or list of input texts.
            model (str, optional): The model to use. Defaults to None.
            auto_retry (bool, optional): Whether to automatically retry the request. Defaults to True.
            strip_newlines (bool, optional): Whether to strip newlines from the input. Defaults to False.
        """
    @overload
    async def async_create_embeddings(
        self,
        inputs: Union[str, List[str]],
        model: Optional[str] = ...,
        auto_retry: Optional[bool] = ...,
        strip_newlines: Optional[bool] = ...,
        headers: Optional[Dict[str, str]] = ...,
        noproxy_required: Optional[bool] = ...,
        include_metadata: Literal[True] = ...,
        **kwargs,
    ) -> Tuple[List[List[float]], Usage]:
        """
        Creates the embeddings

        Args:
            inputs (Union[str, List[str]]): The input text or list of input texts.
            model (str, optional): The model to use. Defaults to None.
            auto_retry (bool, optional): Whether to automatically retry the request. Defaults to True.
            strip_newlines (bool, optional): Whether to strip newlines from the input. Defaults to False.
        """
        ...
    

    async def async_create_embeddings(
        self,
        inputs: Union[str, List[str]],
        model: Optional[str] = None,
        auto_retry: Optional[bool] = True,
        strip_newlines: Optional[bool] = False,
        headers: Optional[Dict[str, str]] = None,
        noproxy_required: Optional[bool] = False,
        include_metadata: Optional[bool] = False,
        **kwargs,
    ) -> List[List[float]] | Tuple[List[List[float]], Usage]:
        """
        Creates the embeddings

        Args:
            inputs (Union[str, List[str]]): The input text or list of input texts.
            model (str, optional): The model to use. Defaults to None.
            auto_retry (bool, optional): Whether to automatically retry the request. Defaults to True.
            strip_newlines (bool, optional): Whether to strip newlines from the input. Defaults to False.
        """
        from lzo.utils.helpers.batching import split_into_batches
        model = model or 'text-embedding-ada-002'
        client = self.get_client(model = model, noproxy_required = noproxy_required, **kwargs)
        inputs = [inputs] if isinstance(inputs, str) else inputs
        if client.supports_tokenization:
            inputs = await self.atruncate_batch_to_max_length(
                inputs, 
                model = model, 
                **kwargs
            )
        
        if strip_newlines: inputs = [i.replace('\n', ' ').strip() for i in inputs]
        
        if not client.is_azure:
            response = await client.embeddings.async_create(input = inputs, model = model, auto_retry = auto_retry, headers = headers, **kwargs)
            return (response.embeddings, response.usage) if include_metadata else response.embeddings

        embeddings = []
        usage = None
        # We need to split into batches of 5 for Azure
        # Azure has a limit of 5 inputs per request
        batches = split_into_batches(inputs, 5)
        for batch in batches:
            response = await client.embeddings.async_create(input = batch, model = model, auto_retry = auto_retry, headers = headers, **kwargs)
            embeddings.extend(response.embeddings)
            if usage is None: usage = response.usage
            else: usage.update(response.usage)
            # Shuffle the clients to load balance
            client = self.get_client(model = model, azure_required = True, noproxy_required = noproxy_required, **kwargs)
        return (embeddings, usage) if include_metadata else embeddings

    acreate_embeddings = async_create_embeddings

    def create_rerankings(
        self,
        query: Optional[str] = None,
        documents: Optional[List[str]] = None,
        top_k: Optional[int] = None,
        model: Optional[str] = None,
        client_name: Optional[str] = None,
        return_documents: Optional[bool] = False,
        truncation: Optional[bool] = True,
        strip_newlines: Optional[bool] = True,
        headers: Optional[Dict[str, str]] = None,
        noproxy_required: Optional[bool] = False,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs,
    ) -> RerankingResponse:
        """
        Creates a reranking response for the provided query and documents

        Usage:

        ```python
        >>> result = OpenAI.create_rerankings(
        >>>    query = 'say this is a test',
        >>>    documents = ['say this is a test', 'say this is another test'],
        >>> )
        ```

        **Parameters:**

        :query (string, required): The query to rerank
        
        :documents (list[string], required): The documents to rerank
        
        :top_k (int, optional): The number of documents to return
        
        :return_documents (bool, optional): If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.
        
        :truncation (bool, optional): Whether to truncate the input text to the maximum context length
        
        :model (string, optional): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.

        Returns: `RerankingResponse`
        """
        if not model and not client_name: 
            default = self.ctx.default_for_routes.get('rerankings')
            client_name, model = default.split('/', 1)
        if strip_newlines: documents = [i.replace('\n', ' ').strip() for i in documents]
        client = self.get_client(model = model, client_name = client_name, noproxy_required = noproxy_required, **kwargs)
        return client.rerankings.create(
            query = query,
            documents = documents,
            top_k = top_k,
            return_documents = return_documents,
            truncation = truncation,
            model = model,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            headers = headers,
            **kwargs,
        )
    
    async def async_create_rerankings(
        self,
        query: Optional[str] = None,
        documents: Optional[List[str]] = None,
        top_k: Optional[int] = None,
        model: Optional[str] = None,
        client_name: Optional[str] = None,
        return_documents: Optional[bool] = False,
        truncation: Optional[bool] = True,
        strip_newlines: Optional[bool] = True,
        headers: Optional[Dict[str, str]] = None,
        noproxy_required: Optional[bool] = False,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        **kwargs,
    ) -> RerankingResponse:
        """
        Creates a reranking response for the provided query and documents

        Usage:

        ```python
        >>> result = OpenAI.acreate_rerankings(
        >>>    query = 'say this is a test',
        >>>    documents = ['say this is a test', 'say this is another test'],
        >>> )
        ```

        **Parameters:**

        :query (string, required): The query to rerank
        
        :documents (list[string], required): The documents to rerank
        
        :top_k (int, optional): The number of documents to return
        
        :return_documents (bool, optional): If false, the API will return a list of {"index", "relevance_score"} where "index" refers to the index of a document within the input list., If true, the API will return a list of {"index", "document", "relevance_score"} where "document" is the corresponding document from the input list.
        
        :truncation (bool, optional): Whether to truncate the input text to the maximum context length
        
        :model (string, optional): ID of the model to use. You can use the List models API to see all of your available models, or see our Model overview for descriptions of them.

        Returns: `RerankingResponse`
        """
        if not model and not client_name: 
            default = self.ctx.default_for_routes.get('rerankings')
            client_name, model = default.split('/', 1)
            
        if strip_newlines: documents = [i.replace('\n', ' ').strip() for i in documents]
        client = self.get_client(model = model, client_name = client_name, noproxy_required = noproxy_required, **kwargs)
        return await client.rerankings.acreate(
            query = query,
            documents = documents,
            top_k = top_k,
            return_documents = return_documents,
            truncation = truncation,
            model = model,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            headers = headers,
            **kwargs,
        )
    
    acreate_rerankings = async_create_rerankings

    
    @overload
    def configure(
        self, 
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        
        api_base: Optional[str] = None,
        api_path: Optional[str] = None,
        api_type: Optional[ApiType] = None,
        api_version: Optional[str] = None,
        api_key_path: Optional[pathlib.Path] = None,

        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        app_info: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,
        ignore_errors: Optional[bool] = None,
        disable_retries: Optional[bool] = None,
        max_connections: Optional[int] = None,
        max_keepalive_connections: Optional[int] = None,
        keepalive_expiry: Optional[int] = None,
        custom_headers: Optional[Dict[str, str]] = None,

        on_error: Optional[Callable] = None,
        reset: Optional[bool] = None,
        prioritize: Optional[str] = None,
        # enable_rotating_clients: Optional[bool] = None,
        azure_model_mapping: Optional[Dict[str, str]] = None,

        auto_healthcheck: Optional[bool] = None,
        auto_loadbalance_clients: Optional[bool] = None,
        proxy_config: Optional[Union[Dict[str, Any], pathlib.Path]] = None,
        client_configurations: Optional[Union[Dict[str, Dict[str, Any]], pathlib.Path]] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client.
        
        :param url: The OpenAI API URL              | Env: [`OPENAI_API_URL`]
        :param scheme: The OpenAI API Scheme        | Env: [`OPENAI_API_SCHEME`]
        :param host: The OpenAI API Host            | Env: [`OPENAI_API_HOST`]
        :param port: The OpenAI API Port            | Env: [`OPENAI_API_PORT`]
        :param api_base: The OpenAI API Base        | Env: [`OPENAI_API_BASE`]
        :param api_key: The OpenAI API Key          | Env: [`OPENAI_API_KEY`]
        :param api_path: The OpenAI API Path        | Env: [`OPENAI_API_PATH`]
        :param api_type: The OpenAI API Type        | Env: [`OPENAI_API_TYPE`]
        :param api_version: The OpenAI API Version  | Env: [`OPENAI_API_VERSION`]
        :param api_key_path: The API Key Path       | Env: [`OPENAI_API_KEY_PATH`]
        :param organization: Organization           | Env: [`OPENAI_ORGANIZATION`]
        :param proxies: The OpenAI Proxies          | Env: [`OPENAI_PROXIES`]
        :param timeout: Timeout in Seconds          | Env: [`OPENAI_TIMEOUT`]
        :param max_retries: The OpenAI Max Retries  | Env: [`OPENAI_MAX_RETRIES`]
        :param ignore_errors: Ignore Errors         | Env: [`OPENAI_IGNORE_ERRORS`]
        :param disable_retries: Disable Retries     | Env: [`OPENAI_DISABLE_RETRIES`]
        :param max_connections: Max Connections     | Env: [`OPENAI_MAX_CONNECTIONS`]
        :param max_keepalive_connections: Max Keepalive Connections | Env: [`OPENAI_MAX_KEEPALIVE_CONNECTIONS`]
        :param keepalive_expiry: Keepalive Expiry   | Env: [`OPENAI_KEEPALIVE_EXPIRY`]
        :param custom_headers: Custom Headers       | Env: [`OPENAI_CUSTOM_HEADERS`]
        
        :param on_error: On Error Callback          
        :param kwargs: Additional Keyword Arguments
        """
        ...
    
    @overload
    def configure(
        azure_api_key: Optional[str] = None,
        azure_url: Optional[str] = None,
        azure_scheme: Optional[str] = None,
        azure_host: Optional[str] = None,
        azure_port: Optional[int] = None,
        
        azure_api_base: Optional[str] = None,
        azure_api_path: Optional[str] = None,
        azure_api_type: Optional[ApiType] = None,
        azure_api_version: Optional[str] = None,
        azure_api_key_path: Optional[pathlib.Path] = None,

        azure_organization: Optional[str] = None,
        azure_proxies: Optional[Union[str, Dict]] = None,
        azure_timeout: Optional[int] = None,
        azure_max_retries: Optional[int] = None,
        azure_app_info: Optional[Dict[str, str]] = None,
        azure_debug_enabled: Optional[bool] = None,
        azure_ignore_errors: Optional[bool] = None,
        azure_max_connections: Optional[int] = None,
        azure_max_keepalive_connections: Optional[int] = None,
        azure_keepalive_expiry: Optional[int] = None,
        azure_custom_headers: Optional[Dict[str, str]] = None,

        on_error: Optional[Callable] = None,
        reset: Optional[bool] = None,
        prioritize: Optional[str] = None,
        # enable_rotating_clients: Optional[bool] = None,
        azure_model_mapping: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,

        auto_healthcheck: Optional[bool] = None,
        auto_loadbalance_clients: Optional[bool] = None,
        proxy_config: Optional[Union[Dict[str, Any], pathlib.Path]] = None,
        client_configurations: Optional[Union[Dict[str, Dict[str, Any]], pathlib.Path]] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client for Azure

        :param azure_url: The OpenAI API URL              | Env: [`AZURE_OPENAI_API_URL`]
        :param azure_scheme: The OpenAI API Scheme        | Env: [`AZURE_OPENAI_API_SCHEME`]
        :param azure_host: The OpenAI API Host            | Env: [`AZURE_OPENAI_API_HOST`]
        :param azure_port: The OpenAI API Port            | Env: [`AZURE_OPENAI_API_PORT`]
        :param azure_api_key: The OpenAI API Key          | Env: [`AZURE_OPENAI_API_KEY`]
        :param azure_api_base: The OpenAI API Base        | Env: [`AZURE_OPENAI_API_BASE`]
        :param azure_api_path: The OpenAI API Path        | Env: [`AZURE_OPENAI_API_PATH`]
        :param azure_api_type: The OpenAI API Type        | Env: [`AZURE_OPENAI_API_TYPE`]
        :param azure_api_version: The OpenAI API Version  | Env: [`AZURE_OPENAI_API_VERSION`]
        :param azure_api_key_path: The API Key Path       | Env: [`AZURE_OPENAI_API_KEY_PATH`]
        :param azure_organization: Organization           | Env: [`AZURE_OPENAI_ORGANIZATION`]
        :param azure_proxies: The OpenAI Proxies          | Env: [`AZURE_OPENAI_PROXIES`]
        :param azure_timeout: Timeout in Seconds          | Env: [`AZURE_OPENAI_TIMEOUT`]
        :param azure_max_retries: The OpenAI Max Retries  | Env: [`AZURE_OPENAI_MAX_RETRIES`]
        :param kwargs: Additional Keyword Arguments
        """
        ...


    def configure(
        self, 
        on_error: Optional[Callable] = None,
        prioritize: Optional[str] = None,
        # enable_rotating_clients: Optional[bool] = None,
        azure_model_mapping: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,

        auto_healthcheck: Optional[bool] = None,
        auto_loadbalance_clients: Optional[bool] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client.
        """
        if on_error is not None: self.on_error = on_error
        if prioritize is not None: self.prioritize = prioritize
        if debug_enabled is not None: self.settings.debug_enabled = debug_enabled
        # if enable_rotating_clients is not None: self.enable_rotating_clients = enable_rotating_clients
        if auto_loadbalance_clients is not None: self.auto_loadbalance_clients = auto_loadbalance_clients
        if auto_healthcheck is not None: self.auto_healthcheck = auto_healthcheck
        if azure_model_mapping is not None: 
            self.azure_model_mapping = azure_model_mapping
            for key, val in azure_model_mapping.items():
                self.ctx.add_model(key, val)
        self.settings.configure(auto_loadbalance_clients = auto_loadbalance_clients, auto_healthcheck = auto_healthcheck, **kwargs)



"""
Jinja2 Extension

- Unifies the Jinja2 Environment handle both sync and async
"""
from typing import TYPE_CHECKING
from .base import Environment, autofilters, FileSystemLoader

if TYPE_CHECKING:
    from jinja2.environment import Template


from __future__ import annotations

"""Factory helpers for configuring LazyOps logging.

The underlying implementation remains untouched – the adjustments here make the
intent of each entry-point explicit for generated documentation and downstream
integrations.
"""

import atexit as _atexit
import os
import sys
import threading
import typing as t

from loguru import _defaults
from loguru._logger import Core as _Core

from .base import InterceptHandler, Logger
from .formatters import LoggerFormatter
from .null_logger import NullLogger
from .state import register_logger_module, run_record_patching_hook
from .static import REVERSE_LOGLEVEL_MAPPING
from .utils import extract_module_name

if t.TYPE_CHECKING:
    from pydantic_settings import BaseSettings
    from logging import Handler as LoggingHandler
    from logging import LogRecord

_lock = threading.Lock()
_logger_contexts: t.Dict[str, Logger] = {}

__all__ = [
    "create_global_logger",
    "create_default_logger",
    "change_logger_level",
    "get_logger",
    "logger",
    "default_logger",
    "null_logger",
]

def create_global_logger(
    name: str | None = "lzl",
    level: str | int = "INFO",
    format: t.Callable[["LogRecord"], str] | None = None,
    filter: t.Callable[["LogRecord"], bool] | None = None,
    handlers: t.Sequence["LoggingHandler"] | None = None,
    settings: "BaseSettings" | None = None,
    **kwargs: t.Any,
) -> Logger:
    """Instantiate the shared global Loguru logger used across LazyOps.

    Args:
        name: Registry key for the logger instance.  Defaults to ``"lzl"``.
        level: Log level for newly attached handlers.  Accepts both string and
            numeric values recognised by Loguru.
        format: Optional callable used to format log records.
        filter: Optional predicate used to filter records.
        handlers: Additional :mod:`logging` handlers to bridge into Loguru.
        settings: Optional settings object attached to the logger for runtime
            reference.
        **kwargs: Additional keyword arguments forwarded to
            :meth:`loguru.Logger.add`.

    Returns:
        Logger: The configured Loguru logger instance.
    """
    global _logger_contexts

    # < 0.7.0
    try:
        _logger = Logger(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=True,
            raw=False,
            capture=True,
            patcher=None,
            extra={},
        )
    # >= 0.7.0
    except Exception as e:
        _logger = Logger(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=False,
            raw=False,
            capture=True,
            patchers=[run_record_patching_hook],
            extra={},
        )

    _logger.name = 'lzl'
    _logger.is_global = True
    dev_level = _logger.level(name='DEV', no=19, color="<blue>", icon="@")

    if _defaults.LOGURU_AUTOINIT and sys.stderr: _logger.add(sys.stderr)
    _atexit.register(_logger.remove)

    _logger.remove()
    _logger.add_if_condition('dev', _logger._is_dev_condition)
    import logging
    logging.basicConfig(
        handlers = handlers or [InterceptHandler()],  level = 0
    )

    _logger.add(
        sys.stdout,
        enqueue = True,
        backtrace = True,
        colorize = True,
        level = level,
        format = format if format is not None else LoggerFormatter.default_formatter,
        filter = filter if filter is not None else _logger._filter,
        **kwargs,
    )
    if settings: _logger.settings = settings
    _logger_contexts[name] = _logger
    return _logger


def create_default_logger(
    name: str | None = None,
    level: str | int = "INFO",
    format: t.Callable[["LogRecord"], str] | None = None,
    filter: t.Callable[["LogRecord"], bool] | None = None,
    handlers: t.Sequence['LoggingHandler'] | None = None,
    settings: 'BaseSettings' | None = None,
    **kwargs: t.Any,
) -> Logger:
    """Return a named logger that proxies calls to the global instance.

    Args:
        name: Optional logger namespace.  If omitted the global logger is
            returned.
        level: Minimum log level once the logger is registered.
        format: Formatter applied when a dedicated handler is configured.
        filter: Optional filtering callable.
        handlers: Additional logging handlers to attach.
        settings: Optional settings object attached to the logger for runtime
            reference.
        **kwargs: Extra keyword arguments forwarded to the Loguru ``add`` call.

    Returns:
        Logger: Either the global logger or a namespaced proxy.
    """
    global _logger_contexts
    if name:
        if name.upper() in REVERSE_LOGLEVEL_MAPPING:
            # If name is a level, then set level to name
            level = name
            name = None
        else:
            name = extract_module_name(name)
            
    if name is None: name = 'lzl'
    if name in _logger_contexts:
        return _logger_contexts[name]
    
    with _lock:
        if name == 'lzl':
            return create_global_logger(
                name = name,
                level = level,
                format = format,
                filter = filter,
                handlers = handlers,
                settings = settings,
            )
        
        if isinstance(level, str): level = level.upper()
        _logger = _logger_contexts['lzl']

        if name and format is not None:
            # Add a new handler
            def _filter_func(record: 'LogRecord') -> bool:
                """
                Filter out messages from other modules
                """
                return extract_module_name(record.name) == name
            
            _logger.add(
                sys.stdout,
                enqueue = True,
                backtrace = True,
                colorize = True,
                level = level,
                format = format,
                filter = _filter_func,
                **kwargs,
            )
            return _logger

        *options, extra = _logger._options
        new_logger = Logger(_logger._core, *options, {**extra})
        if name: 
            _logger_contexts[name] = new_logger
            new_logger.name = name
            register_logger_module(name)

        if settings: new_logger.settings = settings
        # if _is_global_logger: new_logger.is_global = True
        return new_logger


def change_logger_level(
    name: str | None = None,
    level: str | int = "INFO",
    verbose: bool = False,
    **kwargs: t.Any,
) -> None:
    """Update the minimum level for the global or named logger."""
    global logger, logger_level
    if isinstance(level, str): level = level.upper()
    # Skip if the level is the same
    if level == logger_level: return
    name = name or 'lzl'
    name = name.split('.')[0]
    logger_level = level
    if name != 'lzl':
        __logger = get_logger(name, logger_level, **kwargs)
    else:
        __logger = logger
    if verbose: __logger.info(f"[{name}] Changing logger level from {logger_level} -> {level}")
    __logger._core.min_level = float(REVERSE_LOGLEVEL_MAPPING[logger_level.upper()])


if os.getenv('LOGGING_DEBUG_ENABLED', os.getenv('DEBUG_ENABLED')) == 'True':
    logger_level = 'DEBUG'
else:
    logger_level: str = os.getenv('LOGGER_LEVEL', 'INFO').upper()


get_logger = create_default_logger
logger = create_default_logger('lzl', level = logger_level)
default_logger = logger
null_logger = NullLogger(name = 'null_logger')



from __future__ import annotations

"""Public typing façade for LazyOps' higher-level APIs."""

from .base import (
    BaseModel,
    BaseSettings,
    RBaseModel,
    Field,
    eproperty,
    ByteSize,
    field_validator,
    get_schema_extra,
    model_validator,
    pre_root_validator,
    root_validator,
    validator,
)
from .common.appenv import AppEnv, get_app_env
from .common.extra import Final, Literal

__all__ = [
    'AppEnv',
    'ByteSize',
    'BaseModel',
    'BaseSettings',
    'Final',
    'Field',
    'Literal',
    'RBaseModel',
    'eproperty',
    'field_validator',
    'get_app_env',
    'get_schema_extra',
    'model_validator',
    'pre_root_validator',
    'root_validator',
    'validator',
]
